from nlu.discovery import Discoverer
from typing import List, Tuple, Optional, Dict, Union
import streamlit as st
import pandas as pd
from nlu.pipe.viz.streamlit_viz.streamlit_utils_OS import StreamlitUtilsOS
from nlu.pipe.viz.streamlit_viz.styles import _set_block_container_style
from nlu.pipe.viz.streamlit_viz.streamlit_viz_tracker import StreamlitVizTracker
from nlu.pipe.viz.streamlit_viz.viz_building_blocks.dep_tree import DepTreeStreamlitBlock
from nlu.pipe.viz.streamlit_viz.viz_building_blocks.classifier import ClassifierStreamlitBlock
from nlu.pipe.viz.streamlit_viz.viz_building_blocks.token_features import TokenFeaturesStreamlitBlock
from nlu.pipe.viz.streamlit_viz.viz_building_blocks.ner import NERStreamlitBlock
from nlu.pipe.viz.streamlit_viz.viz_building_blocks.word_similarity import WordSimilarityStreamlitBlock
from nlu.pipe.viz.streamlit_viz.viz_building_blocks.word_embedding_manifold import WordEmbeddingManifoldStreamlitBlock
from nlu.pipe.viz.streamlit_viz.viz_building_blocks.sentence_embedding_manifold import \
    SentenceEmbeddingManifoldStreamlitBlock
from nlu.pipe.viz.streamlit_viz.viz_building_blocks.entity_embedding_manifold import \
    EntityEmbeddingManifoldStreamlitBlock


class StreamlitVizBlockHandler():
    """Internal API to access any on the Streamlit building blocks. This is part of the Controller of the MVC pattern"""

    @staticmethod
    def viz_streamlit_dashboard(
            pipe,
            # Base Params
            text: Union[str, List[str], pd.DataFrame, pd.Series],
            model_selection: List[str] = [],
            # NER PARAMS
            # default_ner_model2viz:Union[str, List[str]] = 'en.ner.onto.electra.base',
            # SIMILARITY PARAMS
            similarity_texts: Tuple[str, str] = ('I love NLU <3', 'I love Streamlit <3'),
            title: str = 'NLU ❤️ Streamlit - Prototype your NLP startup in 0 lines of code',
            sub_title: str = 'Play with over 1000+ scalable enterprise NLP models',
            side_info: str = None,
            # UI PARAMS
            visualizers: List[str] = (
            "dependency_tree", "ner", "similarity", "token_features", 'classification', 'manifold'),
            show_models_info: bool = True,
            show_model_select: bool = False,
            show_viz_selection: bool = False,
            show_logo: bool = True,
            set_wide_layout_CSS: bool = True,
            show_code_snippets: bool = False,
            model_select_position: str = 'side',  # main or side
            display_infos: bool = True,
            key: str = "NLU_streamlit",
            display_footer: bool = True,
            num_similarity_cols: int = 2,

            # NEW PARAMS
            # MANIfold
            num_manifold_cols: int = 3,
            manifold_algos: List[str] = ('TSNE'),

            # SIMY
            similarity_algos: List[str] = ('COSINE'),
    ) -> None:
        """Visualize either individual building blocks for streamlit or a full UI to experiment and explore models with"""
        StreamlitVizTracker.footer_displayed = not display_footer
        if set_wide_layout_CSS: _set_block_container_style()
        if title: st.title(title)
        if sub_title: st.subheader(sub_title)
        if show_logo: StreamlitVizTracker.show_logo()
        if side_info: st.sidebar.markdown(side_info)
        if isinstance(text, list): text = '\n'.join(text)
        text = st.text_area("Enter text you want to visualize below", text, key=key)
        ner_model_2_viz = pipe.nlu_ref
        if show_model_select:
            show_code_snippets = st.sidebar.checkbox('Generate code snippets', value=show_code_snippets)
            if model_selection == []: model_selection = Discoverer.get_components('ner', include_pipes=True)
            model_selection.sort()
            if model_select_position == 'side':
                if pipe.nlu_ref.split(' ')[0] in model_selection:
                    ner_model_2_viz = st.sidebar.selectbox("Select a NER model_anno_obj.", model_selection,
                                                           index=model_selection.index(pipe.nlu_ref.split(' ')[0]))
                else:
                    ner_model_2_viz = st.sidebar.selectbox("Select a NER model_anno_obj.", model_selection,
                                                           index=model_selection.index('en.ner'))
            else:
                if pipe.nlu_ref.split(' ')[0] in model_selection:
                    ner_model_2_viz = st.selectbox("Select a NER model_anno_obj", model_selection,
                                                   index=model_selection.index(pipe.nlu_ref.split(' ')[0]))
                else:
                    ner_model_2_viz = st.selectbox("Select a NER model_anno_obj.", index=model_selection.index('en.ner'))

        active_visualizers = visualizers
        if show_viz_selection: active_visualizers = st.sidebar.multiselect("Visualizers", options=visualizers,
                                                                           default=visualizers, key=key)

        all_models = ner_model_2_viz + ' en.dep.typed ' if 'dependency_tree' in active_visualizers else ner_model_2_viz
        ner_pipe, tree_pipe = None, None

        for viz in active_visualizers:
            if 'ner' == viz:
                ner_pipe = pipe if pipe.nlu_ref == ner_model_2_viz else StreamlitUtilsOS.get_pipe(ner_model_2_viz)
                NERStreamlitBlock.visualize_ner(ner_pipe, text, generate_code_sample=show_code_snippets,
                                                key=key + '_ner', show_model_select=False, show_text_input=True,
                                                show_logo=False, show_infos=False)
            if 'dependency_tree' == viz:
                tree_pipe = StreamlitUtilsOS.get_pipe(
                    'en.dep.typed')  # if not ValidateVizPipe.viz_tree_satisfied(component_list) else component_list
                DepTreeStreamlitBlock.visualize_dep_tree(tree_pipe, text, generate_code_sample=show_code_snippets,
                                                         key=key + '_dep', show_infos=False, show_logo=False)
            if 'token_features' == viz:
                ner_pipe = pipe if pipe.nlu_ref == ner_model_2_viz else StreamlitUtilsOS.get_pipe(ner_model_2_viz)
                TokenFeaturesStreamlitBlock.visualize_tokens_information(ner_pipe, text,
                                                                         generate_code_sample=show_code_snippets,
                                                                         key=key + '_tok',
                                                                         model_select_position=model_select_position,
                                                                         show_infos=False, show_logo=False, )
            if 'classification' == viz:
                ner_pipe = pipe if pipe.nlu_ref == ner_model_2_viz else StreamlitUtilsOS.get_pipe(ner_model_2_viz)
                ClassifierStreamlitBlock.visualize_classes(ner_pipe, text, generate_code_sample=show_code_snippets,
                                                           key=key + '_class',
                                                           model_select_position=model_select_position,
                                                           show_infos=False, show_logo=False)
            if 'similarity' == viz:
                ner_pipe = pipe if pipe.nlu_ref == ner_model_2_viz else StreamlitUtilsOS.get_pipe(ner_model_2_viz)
                WordSimilarityStreamlitBlock.display_word_similarity(ner_pipe, similarity_texts,
                                                                     generate_code_sample=show_code_snippets,
                                                                     model_select_position=model_select_position,
                                                                     show_infos=False, show_logo=False,
                                                                     num_cols=num_similarity_cols, key=key + '_sim')
            if 'manifold' == viz:
                ner_pipe = pipe if ner_model_2_viz in pipe.nlu_ref.split(' ') else StreamlitUtilsOS.get_pipe(
                    ner_model_2_viz)
                WordEmbeddingManifoldStreamlitBlock.viz_streamlit_word_embed_manifold(ner_pipe, similarity_texts,
                                                                                      generate_code_sample=show_code_snippets,
                                                                                      model_select_position=model_select_position,
                                                                                      show_infos=False, show_logo=False,
                                                                                      num_cols=num_manifold_cols,
                                                                                      key=key + '_mani')

        models_to_display_info_for = []
        if ner_pipe is not None: models_to_display_info_for.append(ner_pipe)
        if tree_pipe is not None: models_to_display_info_for.append(tree_pipe)
        if show_models_info: StreamlitVizTracker.display_model_info(all_models, models_to_display_info_for)
        if display_infos: StreamlitVizTracker.display_footer()

    @staticmethod
    def viz_streamlit_entity_embed_manifold(
            pipe,  # nlu component_list
            default_texts: List[str] = ("Donald Trump likes to visit New York", "Angela Merkel likes to visit Berlin!",
                                        'Peter hates visiting Paris'),
            title: Optional[str] = "Lower dimensional Manifold visualization for Entity embeddings",
            sub_title: Optional[
                str] = "Apply any of the 11 `Manifold` or `Matrix Decomposition` algorithms to reduce the dimensionality of `Entity Embeddings` to `1-D`, `2-D` and `3-D` ",
            default_algos_to_apply: List[str] = ("TSNE", "PCA"),
            target_dimensions: List[int] = (1, 2, 3),
            show_algo_select: bool = True,
            set_wide_layout_CSS: bool = True,
            num_cols: int = 3,
            model_select_position: str = 'side',  # side or main
            key: str = "NLU_streamlit",
            show_infos: bool = True,
            show_logo: bool = True,
            n_jobs: Optional[int] = 3,  # False
    ):
        EntityEmbeddingManifoldStreamlitBlock.viz_streamlit_entity_embed_manifold(
            pipe,
            default_texts,
            title,
            sub_title,
            default_algos_to_apply,
            target_dimensions,
            show_algo_select,
            set_wide_layout_CSS,
            num_cols,
            model_select_position,
            key,
            show_infos,
            show_logo,
            n_jobs,
        )

    @staticmethod
    def viz_streamlit_sentence_embed_manifold(
            pipe,  # nlu component_list
            default_texts: List[str] = (
            "Donald Trump likes to party!", "Angela Merkel likes to party!", 'Peter HATES TO PARTTY!!!! :('),
            title: Optional[str] = "Lower dimensional Manifold visualization for word embeddings",
            sub_title: Optional[
                str] = "Apply any of the 11 `Manifold` or `Matrix Decomposition` algorithms to reduce the dimensionality of `Word Embeddings` to `1-D`, `2-D` and `3-D` ",
            write_raw_pandas: bool = False,
            default_algos_to_apply: List[str] = ("TSNE", "PCA"),
            # ,'LLE','Spectral Embedding','MDS','ISOMAP','SVD aka LSA','DictionaryLearning','FactorAnalysis','FastICA','KernelPCA',),  # LatentDirichletAllocation 'NMF',
            target_dimensions: List[int] = (1, 2, 3),
            show_algo_select: bool = True,
            show_embed_select: bool = True,
            show_color_select: bool = True,
            MAX_DISPLAY_NUM: int = 100,
            display_embed_information: bool = True,
            set_wide_layout_CSS: bool = True,
            num_cols: int = 3,
            model_select_position: str = 'side',  # side or main
            key: str = "NLU_streamlit",
            additional_classifiers_for_coloring: List[str] = ['pos', 'sentiment'],
            generate_code_sample: bool = False,
            show_infos: bool = True,
            show_logo: bool = True,
            n_jobs: Optional[int] = 3,  # False
    ):
        SentenceEmbeddingManifoldStreamlitBlock.viz_streamlit_sentence_embed_manifold(
            pipe,
            default_texts,
            title,
            sub_title,
            write_raw_pandas,
            default_algos_to_apply,
            target_dimensions,
            show_algo_select,
            show_embed_select,
            show_color_select,
            MAX_DISPLAY_NUM,
            display_embed_information,
            set_wide_layout_CSS,
            num_cols,
            model_select_position,
            key,
            additional_classifiers_for_coloring,
            generate_code_sample,
            show_infos,
            show_logo,
            n_jobs,
        )

    @staticmethod
    def viz_streamlit_word_embed_manifold(
            pipe,  # nlu component_list
            default_texts: List[str] = (
            "Donald Trump likes to party!", "Angela Merkel likes to party!", 'Peter HATES TO PARTTY!!!! :('),
            title: Optional[str] = "Lower dimensional Manifold visualization for word embeddings",
            sub_title: Optional[
                str] = "Apply any of the 11 `Manifold` or `Matrix Decomposition` algorithms to reduce the dimensionality of `Word Embeddings` to `1-D`, `2-D` and `3-D` ",
            write_raw_pandas: bool = False,
            default_algos_to_apply: List[str] = ("TSNE", "PCA"),
            # ,'LLE','Spectral Embedding','MDS','ISOMAP','SVD aka LSA','DictionaryLearning','FactorAnalysis','FastICA','KernelPCA',),  # LatentDirichletAllocation 'NMF',
            target_dimensions: List[int] = (1, 2, 3),
            show_algo_select: bool = True,
            show_embed_select: bool = True,
            show_color_select: bool = True,
            MAX_DISPLAY_NUM: int = 100,
            display_embed_information: bool = True,
            set_wide_layout_CSS: bool = True,
            num_cols: int = 3,
            model_select_position: str = 'side',  # side or main
            key: str = "NLU_streamlit",
            additional_classifiers_for_coloring: List[str] = ['pos', 'sentiment'],
            generate_code_sample: bool = False,
            show_infos: bool = True,
            show_logo: bool = True,
            n_jobs: Optional[int] = 3,  # False
    ):
        WordEmbeddingManifoldStreamlitBlock.viz_streamlit_word_embed_manifold(
            pipe,
            default_texts,
            title,
            sub_title,
            write_raw_pandas,
            default_algos_to_apply,
            target_dimensions,
            show_algo_select,
            show_embed_select,
            show_color_select,
            MAX_DISPLAY_NUM,
            display_embed_information,
            set_wide_layout_CSS,
            num_cols,
            model_select_position,
            key,
            additional_classifiers_for_coloring,
            generate_code_sample,
            show_infos,
            show_logo,
            n_jobs,
        )

    @staticmethod
    def visualize_dep_tree(
            pipe,  # nlu component_list
            text: str = 'Billy likes to swim',
            title: Optional[str] = "Dependency Parse & Part-of-speech tags",
            sub_title: Optional[
                str] = 'POS tags define a `grammatical label` for `each token` and the `Dependency Tree` classifies `Relations between the tokens` ',
            set_wide_layout_CSS: bool = True,
            generate_code_sample: bool = False,
            key="NLU_streamlit",
            show_infos: bool = True,
            show_logo: bool = True,
            show_text_input: bool = True,
    ):
        DepTreeStreamlitBlock.visualize_dep_tree(pipe,
                                                 text,
                                                 title,
                                                 sub_title,
                                                 set_wide_layout_CSS,
                                                 generate_code_sample,
                                                 key, show_infos,
                                                 show_logo,
                                                 show_text_input, )

    @staticmethod
    def display_word_similarity(
            pipe,  # nlu component_list
            default_texts: Tuple[str, str] = ("Donald Trump likes to party!", "Angela Merkel likes to party!"),
            threshold: float = 0.5,
            title: Optional[str] = "Embeddings Similarity Matrix &  Visualizations  ",
            sub_tile: Optional[
                str] = "Visualize `word-wise similarity matrix` and calculate `similarity scores` for `2 texts` and every `word embedding` loaded",
            write_raw_pandas: bool = False,
            display_embed_information: bool = True,
            similarity_matrix=True,
            show_algo_select: bool = True,
            dist_metrics: List[str] = ('cosine'),
            set_wide_layout_CSS: bool = True,
            generate_code_sample: bool = False,
            key: str = "NLU_streamlit",
            num_cols: int = 2,
            display_scalar_similarities: bool = False,
            display_similarity_summary: bool = False,
            model_select_position: str = 'side',  # main or side
            show_infos: bool = True,
            show_logo: bool = True,
    ):
        WordSimilarityStreamlitBlock.display_word_similarity(pipe,
                                                             default_texts,
                                                             threshold,
                                                             title,
                                                             sub_tile,
                                                             write_raw_pandas,
                                                             display_embed_information,
                                                             similarity_matrix,
                                                             show_algo_select,
                                                             dist_metrics,
                                                             set_wide_layout_CSS,
                                                             generate_code_sample,
                                                             key,
                                                             num_cols,
                                                             display_scalar_similarities,
                                                             display_similarity_summary,
                                                             model_select_position,
                                                             show_infos,
                                                             show_logo, )

    @staticmethod
    def visualize_tokens_information(
            pipe,  # nlu component_list
            text: str,
            title: Optional[str] = "Token Features",
            sub_title: Optional[str] = 'Pick from `over 1000+ models` on the left and `view the generated features`',
            show_feature_select: bool = True,
            features: Optional[List[str]] = None,
            full_metadata: bool = True,
            output_level: str = 'token',
            positions: bool = False,
            set_wide_layout_CSS: bool = True,
            generate_code_sample: bool = False,
            key="NLU_streamlit",
            show_model_select=True,
            model_select_position: str = 'side',  # main or side
            show_infos: bool = True,
            show_logo: bool = True,
            show_text_input: bool = True,
    ) -> None:
        TokenFeaturesStreamlitBlock.visualize_tokens_information(
            pipe,
            text,
            title,
            sub_title,
            show_feature_select,
            features,
            full_metadata,
            output_level,
            positions,
            set_wide_layout_CSS,
            generate_code_sample,
            key,
            show_model_select,
            model_select_position,
            show_infos,
            show_logo,
            show_text_input,
        )

    @staticmethod
    def visualize_ner(
            pipe,  # Nlu component_list
            text: str,
            ner_tags: Optional[List[str]] = None,
            show_label_select: bool = True,
            show_table: bool = False,
            title: Optional[str] = "Named Entities",
            sub_title: Optional[
                str] = "Recognize various `Named Entities (NER)` in text entered and filter them. You can select from over `100 languages` in the dropdown.",
            colors: Dict[str, str] = {},
            show_color_selector: bool = False,
            set_wide_layout_CSS: bool = True,
            generate_code_sample: bool = False,
            key="NLU_streamlit",
            model_select_position: str = 'side',
            show_model_select: bool = True,
            show_text_input: bool = True,
            show_infos: bool = True,
            show_logo: bool = True,

    ):
        NERStreamlitBlock.visualize_ner(
            pipe,
            text,
            ner_tags,
            show_label_select,
            show_table,
            title,
            sub_title,
            colors,
            show_color_selector,
            set_wide_layout_CSS,
            generate_code_sample,
            key,
            model_select_position,
            show_model_select,
            show_text_input,
            show_infos,
            show_logo,
        )

    @staticmethod
    def visualize_classes(
            pipe,  # nlu component_list
            text: Union[str, list, pd.DataFrame, pd.Series, List[str]] = (
            'I love NLU and Streamlit and sunny days!', 'I hate rainy daiys', 'CALL NOW AND WIN 1000$M'),
            output_level: Optional[str] = 'document',
            title: Optional[str] = "Text Classification",
            sub_title: Optional[
                str] = 'View predicted `classes` and `confidences` for `hundreds of text classifiers` in `over 200 languages`',
            metadata: bool = False,
            positions: bool = False,
            set_wide_layout_CSS: bool = True,
            generate_code_sample: bool = False,
            key: str = "NLU_streamlit",
            show_model_selector: bool = True,
            model_select_position: str = 'side',
            show_infos: bool = True,
            show_logo: bool = True,
    ) -> None:
        ClassifierStreamlitBlock.visualize_classes(
            pipe,
            text,
            output_level,
            title,
            sub_title,
            metadata,
            positions,
            set_wide_layout_CSS,
            generate_code_sample,
            key,
            show_model_selector,
            model_select_position,
            show_infos,
            show_logo,
        )



class ImageTable2Cell2TextTable:
    @staticmethod
    def get_default_model():
        from sparkocr.transformers import ImageCellsToTextTable
        return ImageCellsToTextTable() \
            .setInputCol("image_region") \
            .setCellsCol('ocr_table_cells')\
            .setOutputCol("ocr_table")




from sparknlp.annotator import SpanBertCorefModel


class CorefBert:
    @staticmethod
    def get_default_model():
        return SpanBertCorefModel.pretrained() \
            .setInputCols(["document", "token"]) \
            .setOutputCol("coref")

    @staticmethod
    def get_pretrained_model(name, language, bucket=None):
        return SpanBertCorefModel.pretrained(name, language, bucket) \
            .setInputCols(["document", "token"]) \
            .setOutputCol("c oref")



'''
Contains methods used to resolve a NLU reference to a NLU component_to_resolve.
Handler for getting default components, etc.
'''
from typing import Dict, List, Union, Optional, Callable, Tuple

from pyspark.ml import PipelineModel, Pipeline
from sparknlp.pretrained import PretrainedPipeline, LightPipeline

from nlu.pipe.nlu_component import NluComponent
from nlu.pipe.utils.component_utils import ComponentUtils
from nlu.pipe.utils.pipe_utils import PipeUtils
from nlu.pipe.utils.resolution.storage_ref_resolution_utils import *
from nlu.spellbook import Spellbook
from nlu.universe.atoms import LicenseType
from nlu.universe.component_universes import ComponentUniverse, anno_class_to_empty_component
from nlu.universe.feature_resolutions import FeatureResolutions
from nlu.universe.feature_universes import NLP_HC_FEATURES, OCR_FEATURES
from nlu.universe.universes import Licenses, license_to_bucket, ModelBuckets

logger = logging.getLogger('nlu')


def init_component(component):
    # Init partial constructor
    if isinstance(component, Callable):
        component = component()
    return component


def is_produced_by_multi_output_component(missing_feature_type: Union[NLP_FEATURES, OCR_FEATURES, NLP_HC_FEATURES]):
    """For these components we resolve to None,
    because they are already beeign satisfied by another component that outputs multiple features
    including the ones coverd here """
    return missing_feature_type == NLP_FEATURES.DOCUMENT_QUESTION_CONTEXT


def resolve_feature(missing_feature_type: Union[NLP_FEATURES, OCR_FEATURES, NLP_HC_FEATURES], language='en',
                    is_licensed=False,
                    is_trainable_pipe=False) -> NluComponent:
    '''
    This function returns a default component_to_resolve for a missing component_to_resolve type
     and core part to the pipeline feature resolution.
    It is used to auto complete pipelines, which are missing required components.
    :param missing_feature_type: String which is either just the component_to_resolve
    type or componenttype@spark_nlp_reference which stems from a models storageref and refers to some pretrained
    embeddings or model_anno_obj
    :return: a NLU component_to_resolve which is a either the default if there is no '@' in the @param
    missing_component_type or a default component_to_resolve for that particular type
    '''
    logger.info(f'Getting default for missing_feature_type={missing_feature_type}')
    if is_produced_by_multi_output_component(missing_feature_type):
        # Special Edge Case for QA annotators
        return None
    if '@' not in missing_feature_type:
        # Resolve feature which has no storage ref or if storage ref is irrelevant at this point
        if is_licensed and is_trainable_pipe and missing_feature_type in FeatureResolutions.default_HC_train_resolutions.keys():
            feature_resolution = FeatureResolutions.default_HC_resolutions[missing_feature_type]
            license_type = Licenses.hc
            model_bucket = ModelBuckets.hc
        elif is_licensed and missing_feature_type in FeatureResolutions.default_HC_resolutions.keys():
            feature_resolution = FeatureResolutions.default_HC_resolutions[missing_feature_type]
            license_type = Licenses.hc
            model_bucket = ModelBuckets.hc
        elif is_licensed and missing_feature_type in FeatureResolutions.default_OCR_resolutions.keys():
            feature_resolution = FeatureResolutions.default_OCR_resolutions[missing_feature_type]
            license_type = Licenses.ocr
            # model_bucket = 'clinical/models' # no bucket based models supported
            model_bucket = ModelBuckets.ocr
        elif missing_feature_type in FeatureResolutions.default_OS_resolutions.keys():
            feature_resolution = FeatureResolutions.default_OS_resolutions[missing_feature_type]
            license_type = Licenses.open_source
            model_bucket = ModelBuckets.open_source
        else:
            raise ValueError(f"Could not resolve feature={missing_feature_type}")
        nlu_component = init_component(feature_resolution.nlu_component)  # Call the partial and init the nlu component

        # Either call get_pretrained(nlp_ref, lang,bucket) or get_default_model() to instantiate Annotator object
        if feature_resolution.get_pretrained:
            return nlu_component.set_metadata(
                nlu_component.get_pretrained_model(feature_resolution.nlp_ref, feature_resolution.language,
                                                   model_bucket),
                feature_resolution.nlu_ref, feature_resolution.nlp_ref, language, False, license_type)
        else:
            return nlu_component.set_metadata(nlu_component.get_default_model(),
                                              feature_resolution.nlu_ref, feature_resolution.nlp_ref, language, False,
                                              license_type)

    else:
        # if there is an @ in the name, we must get some specific
        # pretrained model_anno_obj from the sparknlp reference that should follow after the @
        missing_feature_type, storage_ref = missing_feature_type.split('@')

        if storage_ref == '':
            # Storage ref empty for trainable resolution.
            # Use default embed defined in feature resolution
            if is_licensed and is_trainable_pipe and missing_feature_type in FeatureResolutions.default_HC_train_resolutions.keys():
                feature_resolution = FeatureResolutions.default_HC_resolutions[missing_feature_type]
                license_type = Licenses.hc
                model_bucket = ModelBuckets.hc
            elif missing_feature_type in FeatureResolutions.default_OS_resolutions.keys():
                feature_resolution = FeatureResolutions.default_OS_resolutions[missing_feature_type]
                license_type = Licenses.open_source
                model_bucket = ModelBuckets.open_source
            else:
                raise ValueError(
                    f"Could not resolve empty storage ref with default feature for missing feature = {missing_feature_type}")
            nlu_component = init_component(
                feature_resolution.nlu_component)  # Call the partial and init the nlu component
            return nlu_component.set_metadata(
                nlu_component.get_pretrained_model(feature_resolution.nlp_ref, feature_resolution.language,
                                                   model_bucket),
                feature_resolution.nlu_ref, feature_resolution.nlp_ref, language, False, license_type)

        # Actually resolve storage ref
        nlu_ref, nlp_ref, is_licensed, language = resolve_storage_ref(language, storage_ref, missing_feature_type)
        license_type = Licenses.hc if is_licensed else Licenses.open_source
        nlu_component = get_trained_component_for_nlp_model_ref(language, nlu_ref, nlp_ref, license_type)
        return nlu_component


def nlu_ref_to_component(nlu_ref, detect_lang=False, authenticated=False) -> Union[NluComponent, List[NluComponent]]:
    '''
    This method implements the main namespace for all component_to_resolve names. It parses the input request and passes
    the data to a resolver method which searches the namespace for a Component for the input request
    It returns a list of NLU.component_to_resolve objects or just one NLU.component_to_resolve
    object alone if just one component_to_resolve was specified.
    It maps a correctly namespaced name to a corresponding component_to_resolve for pipeline
    If no lang is provided, default language eng is assumed.
    General format  <lang>.<class>.<dataset>.<embeddings>
    For embedding format : <lang>.<class>.<variant>
    This method will parse <language>.<NLU_action>
        Additional data about dataset and variant will be resolved by corrosponding action classes

    If train prefix is part of the nlu_ref ,the trainable namespace will e searched

    if 'translate_to' or 'marian' is inside the nlu_ref, 'xx' will be prefixed to the ref and set as lang if it is not already
    Since all translate models are xx lang
    :param nlu_ref: User request (should be a NLU reference)
    :param detect_lang: Whether to automatically  detect language
    :return: Pipeline or component_to_resolve for the NLU reference.
    '''

    infos = nlu_ref.split('.')
    if len(infos) == 0:
        raise ValueError(f"EXCEPTION: Could not create a component_to_resolve for nlu reference={nlu_ref}", )

    if 'train' in infos:
        if nlu_ref in Spellbook.trainable_models.keys():
            if nlu_ref not in Spellbook.trainable_models:
                s = "\n"
                raise ValueError(f'Could not find trainable model_anno_obj for nlu_ref={nlu_ref}.'
                                 f'Supported values = {s.join(nlu.Spellbook.trainable_models.keys())}')
            return get_trainable_component_for_nlu_ref(nlu_ref)
    lang, nlu_ref, nlp_ref, license_type, is_pipe, model_params = nlu_ref_to_nlp_metadata(nlu_ref)

    if is_pipe:
        resolved_component, _ = get_trained_component_list_for_nlp_pipe_ref(lang, nlp_ref, nlu_ref,
                                                                         license_type=license_type)
    else:
        resolved_component = get_trained_component_for_nlp_model_ref(lang, nlu_ref, nlp_ref, license_type, model_params)

    if resolved_component is None:
        raise ValueError(f"EXCEPTION: Could not create a component_to_resolve for nlu reference={nlu_ref}", )
    return resolved_component


def get_trainable_component_for_nlu_ref(nlu_ref) -> NluComponent:
    if nlu_ref in Spellbook.traianble_nlu_ref_to_jsl_anno_id:
        anno_id = Spellbook.traianble_nlu_ref_to_jsl_anno_id[nlu_ref]
    else:
        raise ValueError(f'Could not find trainable Model for nlu_spell ={nlu_ref}')
    if anno_id in ComponentUniverse.components:
        component = ComponentUniverse.components[anno_id]()
        return component.set_metadata(component.get_trainable_model(), nlu_ref, '', 'xx', False)
    else:
        raise ValueError(f'Could not find trainable Model for anno_id ={anno_id}')


def get_trained_component_list_for_nlp_pipe_ref(language, nlp_ref, nlu_ref, path=None,
                                                      license_type: LicenseType = Licenses.open_source,
                                                      ) -> Tuple[List[NluComponent],str]:
    """
    creates a list of components from a Spark NLP Pipeline reference
    1. download pipeline
    2. unpack pipeline to annotators and create list of nlu components
    3. return list of nlu components
    :param license_type: Type of license for the component
    :param nlu_ref: Nlu ref that points to this pipe
    :param language: language of the pipeline
    :param nlp_ref: Reference to a spark nlp pretrained pipeline
    :param path: Load component_list from HDD
    :return: Tuple,
                first element List of each element of the Spark NLP pipeline wrapped as a NLU component_to_resolve inside a list
                second element UUID of the pipeline
    """
    logger.info(f'Building pretrained pipe for nlu_ref={nlu_ref} nlp_ref={nlp_ref}')
    if 'language' in nlp_ref:
        # special edge case for lang detectors
        language = 'xx'
    if path is None:
        if license_type != Licenses.open_source:
            pipe = PretrainedPipeline(nlp_ref, lang=language, remote_loc='clinical/models')
            uid = pipe.model.uid

        else:
            pipe = PretrainedPipeline(nlp_ref, lang=language)
            uid = pipe.model.uid
        iterable_stages = pipe.light_model.pipeline_model.stages
    else:
        pipe = LightPipeline(PipelineModel.load(path=path))
        uid = pipe.pipeline_model.uid
        iterable_stages = pipe.pipeline_model.stages
    constructed_components = get_component_list_for_iterable_stages(iterable_stages, language, nlp_ref, nlu_ref,
                                                                    license_type)
    return ComponentUtils.set_storage_ref_attribute_of_embedding_converters(
        PipeUtils.set_column_values_on_components_from_pretrained_pipe(constructed_components)), uid



def get_nlu_pipe_for_nlp_pipe(pipe: Union[Pipeline, LightPipeline, PipelineModel, List], is_pre_configured=True):
    """Get a list of NLU components wrapping each Annotator in pipe.
    Pipe should be of class Pipeline, LightPipeline, or PipelineModel
    :param pipe: for which to extract list of nlu components which embellish each annotator
    :return: list of nlu components, one per annotator in pipe
    """

    if isinstance(pipe, List):
        pipe = get_component_list_for_iterable_stages(pipe, is_pre_configured=is_pre_configured)
    elif isinstance(pipe, Pipeline):
        pipe = get_component_list_for_iterable_stages(pipe.getStages(), is_pre_configured=is_pre_configured)
    elif isinstance(pipe, LightPipeline):
        pipe = get_component_list_for_iterable_stages(pipe.pipeline_model.stages, is_pre_configured=is_pre_configured)
    elif isinstance(pipe, PipelineModel):
        pipe = get_component_list_for_iterable_stages(pipe.stages, is_pre_configured=is_pre_configured)
    elif isinstance(pipe, PretrainedPipeline):
        pipe = get_component_list_for_iterable_stages(pipe.model.stages, is_pre_configured=is_pre_configured)
    else:
        raise ValueError(
            f'Invalid Pipe-Like class {type(pipe)} supported types: Pipeline,LightPipeline,PipelineModel,List')
    if is_pre_configured:
        return set_cols_on_nlu_components(pipe)
    else:
        return pipe


def set_cols_on_nlu_components(iterable_components):
    for c in iterable_components:
        c.spark_input_column_names = c.model.getInputCols() if hasattr(c.model, 'getInputCols') else [
            c.model.getInputCol()]
        if hasattr(c.model, 'getOutputCol'):
            c.spark_output_column_names = [c.model.getOutputCol()]
        elif hasattr(c.model, 'getOutputCols'):
            c.spark_output_column_names = [c.model.getOutputCols()]

    return iterable_components


def get_component_list_for_iterable_stages(iterable_stages, language=None, nlp_ref=None, nlu_ref=None,
                                           is_pre_configured=True
                                           ):
    constructed_components = []
    for jsl_anno_object in iterable_stages:
        anno_class_name = type(jsl_anno_object).__name__
        logger.info(f"Building NLU component for class_name = {anno_class_name} ")
        component = anno_class_to_empty_component(anno_class_name)

        component.set_metadata(jsl_anno_object, nlu_ref, nlp_ref, language, is_pre_configured)
        constructed_components.append(component)
        if None in constructed_components or len(constructed_components) == 0:
            raise Exception(f"Failure inferring type anno_class={anno_class_name} ")
    return constructed_components


def get_trained_component_for_nlp_model_ref(lang: str, nlu_ref: Optional[str] = '', nlp_ref: str = '',
                                            license_type: LicenseType = Licenses.open_source,
                                            model_configs: Optional[Dict[str, any]] = None) -> NluComponent:
    anno_class = Spellbook.nlp_ref_to_anno_class[nlp_ref]
    component = anno_class_to_empty_component(anno_class)
    model_bucket = license_to_bucket(license_type)
    try:
        if component.get_pretrained_model:
            component = component.set_metadata(
                component.get_pretrained_model(nlp_ref, lang, model_bucket),
                nlu_ref, nlp_ref, lang, False, license_type)
        else:
            component = component.set_metadata(component.get_default_model(),
                                               nlu_ref, nlp_ref, lang, False, license_type)
        if model_configs:
            for method_name, parameter in model_configs.items():
                # Dynamically call method from provided name and value, to set parameters like T5 task
                code = f'component.model.{method_name}({parameter})'
                eval(code)
    except Exception as e:
        raise ValueError(f'Failure making component, nlp_ref={nlp_ref}, nlu_ref={nlu_ref}, lang={lang}, \n err={e}')

    return component



class SpanMedical:
    @staticmethod
    def get_default_model():
        from sparknlp_jsl.annotator import MedicalQuestionAnswering

        return MedicalQuestionAnswering.pretrained() \
            .setInputCols(["document_question", "context"]) \
            .setOutputCol("answer")



    @staticmethod
    def get_pretrained_model(name, language, bucket=None):
        from sparknlp_jsl.annotator import MedicalQuestionAnswering

        return MedicalQuestionAnswering.pretrained(name, language, bucket) \
            .setInputCols(["document_question", "context"]) \
            .setOutputCol("answer")



import json
import os

from nlu.universe.annotator_class_universe import AnnoClassRef
from nlu.universe.component_universes import  jsl_id_to_empty_component
from nlu.universe.universes import Licenses

# These Spark NLP imports are not unused and will be dynamically loaded, do not remove!
from sparknlp.annotator import *
from sparknlp.base import *
# we just use some classes here, so that intelij cleanup will not remove the imports
DocumentAssembler.name,BertEmbeddings.__name__


def is_pipe(model_path):
    """Check whether there is a pyspark component_list stored in path"""
    return 'stages' in os.listdir(model_path)


def is_model(model_path):
    """Check whether there is a model_anno_obj stored in path"""
    return 'metadata' in os.listdir(model_path)


def get_model_class(model_path):
    """Extract class from a model_anno_obj saved in model_path"""
    with open(model_path + '/stages/part-00000', encoding="utf8") as json_file:
        java_class = json.load(json_file)['class']
        pyth_class = java_class.split('.')[-1]
    return java_class, pyth_class


def verify_and_create_model(model_path: str):
    """
     Build model_anno_obj with requirements
     Figures out class name by checking metadata json file
     assumes metadata is always called part-00000
    """
    with open(model_path + '/metadata/' + 'part-00000', encoding="utf8") as json_f:
        class_name = json.load(json_f)['class'].split('.')[-1]
        # The last element in the Class name can be used to just load the model_anno_obj from disk!
        # Just call eval on it, which will give you the actual Python class reference which should have a .load() method
        try:

            m = eval(class_name).load(model_path)
        except:
            from nlu.utils.environment.offline_load_utils_licensed import verify_model_licensed
            m = verify_model_licensed(class_name, model_path)
    os_annos = AnnoClassRef.get_os_pyclass_2_anno_id_dict()
    hc_annos = AnnoClassRef.get_hc_pyclass_2_anno_id_dict()
    ocr_annos = AnnoClassRef.get_ocr_pyclass_2_anno_id_dict()

    # component_type, nlu_anno_class, = resolve_annotator_class_to_nlu_component_info(class_name)

    # construct_component_from_identifier('xx', nlu_ref = class_name, nlp_ref = class_name, anno_class_name=class_name)
    if class_name in os_annos.keys():
        jsl_anno_id = os_annos[class_name]
        nlu_component = jsl_id_to_empty_component(jsl_anno_id)
        return nlu_component.set_metadata(m,
                                          jsl_anno_id, jsl_anno_id,
                                          'xx',
                                          False, Licenses.open_source)

    elif class_name in hc_annos.keys():
        jsl_anno_id = hc_annos[class_name]
        nlu_component = jsl_id_to_empty_component(jsl_anno_id)
        return nlu_component.set_metadata(m,
                                          jsl_anno_id, jsl_anno_id,
                                          'xx',
                                          False, Licenses.hc)
    elif class_name in ocr_annos.keys():
        pass
    raise ValueError(f'Could not detect correct Class for nlp class ={class_name}')



def test_check_if_string_in_file(file_name, string_to_search, regex=False):
    """ Check if any line in the file contains given string """
    # Open the file in read only mode
    # print('reading ', file_name)
    import re

    with open(file_name, 'r', encoding="utf8") as read_obj:
        # Read all lines in the file one by one
        for line in read_obj:
            # For each line, check if line contains the string
            if regex:
                if len(re.findall(string_to_search, line)) > 0: return True
            else:
                if string_to_search in line: return True
    return False


def NLP_ref_to_NLU_ref(nlp_ref, lang):
    """Resolve a Spark NLP reference to a NLU reference"""
    from nlu import Spellbook
    nlu_namespaces_to_check = [Spellbook.pretrained_pipe_references, Spellbook.pretrained_models_references,
                               Spellbook.pretrained_healthcare_model_references, Spellbook.pretrained_pipe_references,
                               ]
    for dict_ in nlu_namespaces_to_check:
        if lang in dict_.keys():
            for reference in dict_[lang]:
                if dict_[lang][reference] == nlp_ref:
                    return reference



"""
Resolve Annotator Classes in the Pipeline to Extractor Configs and Methods.
Each Spark NLP Annotator Class is mapped to at least one

Every Annotator should have 2 configs. Some might offor multuple configs/method pairs, based on model_anno_obj/NLP reference.
- default/minimalistic -> Just the results of the annotations, no confidences or extra metadata
- with meta            -> A config that leverages white/black list and gets the most relevant metadata
- with positions       -> With Begins/Ends


"""
from sparknlp.annotator import *
from sparknlp.base import *
from nlu.pipe.extractors.extractor_configs_OS import *
from nlu.pipe.col_substitution.col_substitution_OS import *

OS_anno2substitution_fn = {
    NerConverter: {
        'default': substitute_ner_converter_cols,
    },
    PerceptronModel: {
        'default': substitute_pos_cols,
    },
    BertEmbeddings: {
        'default': substitute_word_embed_cols,
    },
    AlbertEmbeddings: {
        'default': substitute_word_embed_cols,
    },
    XlnetEmbeddings: {
        'default': substitute_word_embed_cols,
    },

    DistilBertEmbeddings: {
        'default': substitute_word_embed_cols,
    },
    RoBertaEmbeddings: {
        'default': substitute_word_embed_cols,
    },
    XlmRoBertaEmbeddings: {
        'default': substitute_word_embed_cols,
    },

    WordEmbeddingsModel: {
        'default': substitute_word_embed_cols,
    },
    ElmoEmbeddings: {
        'default': substitute_word_embed_cols,
    },
    E5Embeddings: {
        'default': substitute_word_embed_cols,
    },
    BertSentenceEmbeddings: {
        'default': substitute_sent_embed_cols,
    },
    RoBertaSentenceEmbeddings: {
        'default': substitute_sent_embed_cols,
    },
    InstructorEmbeddings: {
        'default': substitute_sent_embed_cols,
    },

    Doc2VecModel: {
        'default': substitute_sent_embed_cols,
    },

    XlmRoBertaSentenceEmbeddings: {
        'default': substitute_sent_embed_cols,
    },
    UniversalSentenceEncoder: {
        'default': substitute_sent_embed_cols,
    },
    SentenceEmbeddings: {
        'default': substitute_sent_embed_cols,
    },
    MPNetEmbeddings: {
        'default': substitute_sent_embed_cols,
    },
    Tokenizer: {
        'default': substitute_tokenizer_cols,
    },
    TokenizerModel: {
        'default': substitute_tokenizer_cols,
    },
    RegexTokenizer: {
        'default': substitute_tokenizer_cols,
    },
    DocumentAssembler: {
        'default': substitute_doc_assembler_cols,
    },
    SentenceDetectorDLModel: {
        'default': substitute_sentence_detector_dl_cols,
    },
    SentenceDetector: {
        'default': substitute_sentence_detector_pragmatic_cols,
    },
    ContextSpellCheckerModel: {
        'default': substitute_spell_context_cols,
    },
    SymmetricDeleteModel: {
        'default': substitute_spell_symm_cols,
    },
    NorvigSweetingModel: {
        'default': substitute_spell_norvig_cols,
    },
    LemmatizerModel: {
        'default': substitute_lem_cols,
    },
    Normalizer: {
        'default': substitute_norm_cols,
    },
    NormalizerModel: {
        'default': substitute_norm_cols,
    },
    DocumentNormalizer: {
        'default': substitute_doc_norm_cols,
    },
    Stemmer: {
        'default': substitute_stem_cols,
    },
    NerDLModel: {
        'default': substitute_ner_dl_cols,
    },
    NerCrfModel: {
        'default': 'TODO',
    },
    LanguageDetectorDL: {
        'default': 'TODO',
    },
    DependencyParserModel: {
        'default': substitute_un_labled_dependency_cols,
    },
    TypedDependencyParserModel: {
        'default': substitute_labled_dependency_cols,
    },
    SentimentDLModel: {
        'default': substitute_sentiment_dl_cols,
    },
    SentimentDetectorModel: {
        'default': substitute_sentiment_dl_cols,
    },
    ViveknSentimentModel: {
        'default': substitute_sentiment_vivk_cols,
    },
    MultiClassifierDLModel: {
        'default': substitute_multi_classifier_dl_cols,
    },
    ClassifierDLModel: {
        'default': substitute_classifier_dl_cols,
    },

    Chunker: {
        'default': substitute_chunk_cols,
    },
    NGramGenerator: {
        'default': substitute_ngram_cols,
    },
    ChunkEmbeddings: {
        'default': substitute_chunk_embed_cols,
    },
    StopWordsCleaner: {
        'default': substitute_stopwords_cols,
    },

    BertForTokenClassification: {
        'default': substitute_transformer_token_classifier_cols,
    },

    DistilBertForTokenClassification: {
        'default': substitute_transformer_token_classifier_cols,
    },

    XlnetForTokenClassification: {
        'default': substitute_transformer_token_classifier_cols,
    },

    XlmRoBertaForTokenClassification: {
        'default': substitute_transformer_token_classifier_cols,
    },

    RoBertaForTokenClassification: {
        'default': substitute_transformer_token_classifier_cols,
    },

    LongformerForTokenClassification: {
        'default': substitute_transformer_token_classifier_cols,
    },

    AlbertForTokenClassification: {
        'default': substitute_transformer_token_classifier_cols,
    },

    BertForSequenceClassification: {
        'default': substitute_seq_bert_classifier_cols
    },
    DistilBertForSequenceClassification: {
        'default': substitute_seq_bert_classifier_cols
    },
    TextMatcherModel: {
        'default': substitute_text_match_cols,
    },
    RegexMatcherModel: {
        'default': substitute_regex_match_cols,
    },
    DateMatcher: {
        'default': substitute_date_match_cols,
    },
    MultiDateMatcher: {
        'default': '',  # TODO
    },

    Doc2Chunk: {
        'default': substitute_doc2chunk_cols,
    },
    #
    Chunk2Doc: {
        'default': substitute_doc2chunk_cols,  # TODO better?
    },
    T5Transformer: {
        'default': substitute_T5_cols,
    },
    MarianTransformer: {
        'default': substitute_marian_cols,
    },
    YakeKeywordExtraction: {
        'default': substitute_YAKE_cols,
    },
    WordSegmenterModel: {
        'default': substitute_word_seg_cols,
    },


    # approaches
    ViveknSentimentApproach: {'default': substitute_sentiment_vivk_approach_cols,
                              'default_full': default_full_config, },
    SentimentDLApproach: {'default': substitute_sentiment_dl_approach_cols, 'default_full': default_full_config, },
    ClassifierDLApproach: {'default': substitute_classifier_dl_approach_cols, 'default_full': default_full_config, },
    MultiClassifierDLApproach: {'default': substitute_multi_classifier_dl_approach_cols,
                                'default_full': default_full_config, },
    NerDLApproach: {'default': substitute_ner_dl_approach_cols, 'default_full': default_full_config, },
    PerceptronApproach: {'default': substitute_pos_approach_cols, 'default_full': default_full_config, },
    Doc2VecApproach: {'default': substitute_sent_embed_cols},

}





class ZeroShotRelationExtractor:
    @staticmethod
    def get_default_model():
        from sparknlp_jsl.annotator import ZeroShotRelationExtractionModel
        return ZeroShotRelationExtractionModel.pretrained(name = 'redl_bodypart_direction_biobert') \
                   .setInputCols(["entities", "sentence"]) \
                   .setOutputCol("relations")


    @staticmethod
    def get_pretrained_model(name, language, bucket='clinical/models'):
        from sparknlp_jsl.annotator import ZeroShotRelationExtractionModel
        return ZeroShotRelationExtractionModel.pretrained(name, language,bucket) \
            .setInputCols(["entities", "sentence"]) \
            .setOutputCol("relations")




class ZeroShotNer:
    @staticmethod
    def get_default_model():
        from sparknlp_jsl.annotator import ZeroShotNerModel
        return ZeroShotNerModel().setInputCols(["sentence", "token"]) \
            .setOutputCol("zero_shot_ner") \
            .setPredictionThreshold(0.1)

    @staticmethod
    def get_pretrained_model(name, language, bucket=None):
        from sparknlp_jsl.annotator import ZeroShotNerModel
        return ZeroShotNerModel.pretrained(name, language, bucket).setInputCols(["sentence", "token"]) \
            .setOutputCol("zero_shot_ner") \
            .setPredictionThreshold(0.1)



"""
Collection of universes shared across all libraries (NLP/HC/OCR), which are collections of atoms
"""
from nlu.universe.atoms import LicenseType, NlpLevel


class NLP_LEVELS(NlpLevel):
    """
    XXX_SUPER is a N to M Mapping, with M <= N
    XXX_SUB is a N to M mapping, with M >=N
    no prefix implies a N to N mapping to be expected
    """
    DOCUMENT = NlpLevel('document')
    CHUNK = NlpLevel('chunk')
    SENTENCE = NlpLevel('sentence')
    TOKEN = NlpLevel('token')
    CO_REFERENCE = NlpLevel('coreference')
    RELATION = NlpLevel('relation')
    MULTI_TOKEN_CLASSIFIER = NlpLevel('multi_token_classifier')

    AUDIO_SERIES = NlpLevel('audio_series')

    IMAGE= NlpLevel("image")
    INPUT_DEPENDENT_DOCUMENT_CLASSIFIER = NlpLevel('input_dependent_document_classifier')
    INPUT_DEPENDENT_DOCUMENT_EMBEDDING = NlpLevel('input_dependent_document_embedding')

    # Not used for now
    # NGRAM_CHUNK = NlpLevel('NGRAM_CHUNK')
    # SUB_TOKEN = NlpLevel('sub_token')
    # SUPER_TOKEN = NlpLevel('super_token')
    # SUPER_CHUNK = NlpLevel('super_chunk')
    # SUB_CHUNK = NlpLevel('sub_chunk')
    # POS_CHUNK = NlpLevel('POS_CHUNK')
    # KEYWORD_CHUNK = NlpLevel('KEYWORD_CHUNK')
    # NER_CHUNK = NlpLevel("ner_chunk")


class OCR_OUTPUT_LEVELS:
    # PAGES ARE LIKE TOKENS!! Book is full document!

    PAGES = 'pages'  # Generate 1 output per PAGE in each input document. I.e if 2 PDFs input with 5 pages each, gens 10 rows. 1 to many mapping
    FILE = 'file'  # Generate 1 output per document, I.e. 2 PDFS with 5 pages each gen 2 Row, 1 to one mapping
    OBJECT = 'object'  # Generate 1 output row per detected Object in Input document. I.e. if 2 PDFS with 5 Cats each, generates 10 rows. ---> REGION or Not?
    CHARACTER = 'character'  # Generate 1 oputput row per OCR'd character, I.e. 2 PDFS with 100 Chars each, gens 100 Rows.
    TABLE = 'table'  # 1 Pandas DF per Table.


class AnnoTypes:
    # DOCUMENT_XX can be sbustituted for SENTENCE
    CHUNK_MAPPER = 'chunk_mapper'

    TOKENIZER = 'tokenizer'
    TOKEN_CLASSIFIER = 'token_classifier'
    QUESTION_SPAN_CLASSIFIER = 'span_classifier'
    QUESTION_TABLE_ANSWERER = 'question_table_answerer'
    TRANSFORMER_TOKEN_CLASSIFIER = 'transformer_token_classifier'  # Can be token level but also NER level
    TRANSFORMER_SEQUENCE_CLASSIFIER = 'transformer_sequence_classifier'  # Can be token level but also NER level
    CHUNK_CLASSIFIER = 'chunk_classifier'  # ASSERTION/ NER GENERATES/CONTEXT_PARSER THESE but DOES NOT TAKE THEM IN!!! Split into NER-CHUNK Classifier, etc..?
    DOCUMENT_CLASSIFIER = 'document_classifier'
    RELATION_CLASSIFIER = 'relation_classifier'  # Pairs of chunks
    TOKEN_EMBEDDING = 'token_embedding'
    CHUNK_EMBEDDING = 'chunk_embedding'
    DOCUMENT_EMBEDDING = 'document_embedding'
    SENTENCE_DETECTOR = 'sentence_detector'
    SENTENCE_EMBEDDING = 'sentence_embedding'

    SPELL_CHECKER = 'spell_checker'
    HELPER_ANNO = 'helper_anno'
    SPEECH_RECOGNIZER = 'speech_recognizer'
    TEXT_NORMALIZER = 'text_normalizer'
    TOKEN_NORMALIZER = 'token_normalizer'
    # TODO chunk sub-classes? I.e. POS-CHUNKS, NER-CHUNKS, KEYWORD-CHUNKS, RESOLUTION-CHUNKS, etc??
    pos_regex_chunker = 'token_normalizer'
    CHUNK_FILTERER = 'chunk_filterer'

    TEXT_RECOGNIZER = 'text_recognizer'
    TABLE_RECOGNIZER = 'table_recognizer'
    PDF_BUILDER = 'table_recognizer'
    OCR_UTIL = 'ocr_util'
    IMAGE_CLASSIFICATION = 'image_classifier'

    PARTIALLY_READY = 'partially_ready'



from sparknlp.annotator import Word2VecModel, Word2VecApproach


class Word2Vec:
    @staticmethod
    def get_default_model():
        return Word2VecModel.pretrained() \
        .setInputCols("token") \
        .setOutputCol("word_embeddings")

    @staticmethod
    def get_pretrained_model(name, language, bucket=None):
        return Word2VecModel.pretrained(name,language,bucket) \
            .setInputCols("token") \
            .setOutputCol("word_embeddings")
    @staticmethod
    def get_trainable_model():
        return Word2VecApproach()\
            .setInputCols("token") \
            .setOutputCol("word_embeddings")





class Img2Text:
    @staticmethod
    def get_default_model():
        from sparkocr.transformers import ImageToText
        return ImageToText() \
            .setInputCol("image") \
            .setOutputCol("text") \
            .setOcrParams(["preserve_interword_spaces=1", ])




from typing import List

import pandas as pd


def extract_table(df):
    import pyspark.sql.functions as f
    columns = df.select(f.size(f.col("ocr_table.chunks").getItem(0).chunkText)).collect()[0][0]
    exploded_results = df \
        .select("ocr_table") \
        .withColumn("cells", f.explode(f.col("ocr_table.chunks"))) \
        .select([f.col("cells")[i].getField("chunkText").alias(f"col{i}") for i in
                 range(0, columns)])
    return exploded_results.toPandas()


def extract_tables(df, rename_cols=False):
    df = df.withColumn("table_index", df.ocr_table.area.index)
    # pagennum
    pandas_tables = []
    for r in df.dropDuplicates(["path", "ocr_table", "table_index"]).collect():
        if r.table_index is not None:
            pandas_tables.append(extract_table(df.filter(
                (df.path == r.path) & (df.ocr_table.area.page == r.ocr_table.area.page) & (
                        df.table_index == r.table_index))))
    if rename_cols:
        pandas_tables = use_first_row_as_column_names_for_list_of_dfs(pandas_tables)

    pandas_tables = rename_duplicate_cols(pandas_tables)
    return pandas_tables


def rename_duplicate_cols(dfs: List[pd.DataFrame]):
    for df in dfs:
        import collections
        duplicates = [item for item, count in collections.Counter(df.columns).items() if count > 1]
        new_cols = []

        for i, c in enumerate(df.columns):
            if c in duplicates:
                # update duplicate cols
                c = f'{i}_{c}'
            new_cols.append(c)
        df.columns = new_cols
    return dfs


def use_first_row_as_column_names(df):
    # Use first row of df to define columns and drop the row afterwards
    if len(df) > 1:
        df.columns = df.iloc[0].values
        df = df.iloc[1:]
    return df


def use_first_row_as_column_names_for_list_of_dfs(pd_tables):
    new_tables = []
    for t in pd_tables:
        new_tables.append(use_first_row_as_column_names(t))
    return new_tables



import logging
import nlu
import requests

logger = logging.getLogger('nlu')


class ModelHubUtils():
    """Modelhub utils"""
    modelhub_json_url = 'https://nlp.johnsnowlabs.com/models.json'
    data = requests.get(modelhub_json_url).json()

    @staticmethod
    def NLU_ref_to_NLP_ref(nlu_ref: str, lang: str = None) -> str:
        """Resolve a Spark NLU reference to a NLP reference.
        :param nlu_ref: which nlu model_anno_obj's nlp refrence to return.
        :param lang: what language is the model_anno_obj in.
        :return: Spark nlp model_anno_obj name
        """
        nlu_namespaces_to_check = [nlu.Spellbook.pretrained_pipe_references, nlu.Spellbook.pretrained_models_references,
                                   nlu.Spellbook.pretrained_healthcare_model_references,
                                   nlu.Spellbook.licensed_storage_ref_2_nlu_ref,
                                   nlu.Spellbook.storage_ref_2_nlu_ref]  # ]
        for dict_ in nlu_namespaces_to_check:
            if lang:
                if lang in dict_.keys():
                    for reference in dict_[lang]:
                        if reference == nlu_ref:
                            return dict_[lang][reference]
            else:
                for dict_ in nlu_namespaces_to_check:
                    for lang in dict_:
                        for reference in dict_[lang]:
                            if reference == nlu_ref:
                                return dict_[lang][reference]
        for _nlp_ref, nlp_ref_type in nlu.Spellbook.component_alias_references.items():
            if _nlp_ref == nlu_ref: return nlp_ref_type[0]
        return ''

    @staticmethod
    def get_url_by_nlu_refrence(nlu_ref: str) -> str:
        """Resolves a URL for an NLU reference.
        :param nlu_ref: Which nlu refrence's url to return.
        :return: url to modelhub
        """

        # getting spark refrence for given nlu refrence
        if not nlu_ref: return 'https://nlp.johnsnowlabs.com/models'
        if nlu_ref.split(".")[0] not in nlu.Spellbook.pretrained_models_references.keys():
            nlu_ref = "en." + nlu_ref
        nlp_refrence = ModelHubUtils.NLU_ref_to_NLP_ref(nlu_ref)
        if nlp_refrence == None:
            print(f"{nlp_refrence}         {nlu_ref}")
            return 'https://nlp.johnsnowlabs.com/models'
        else:
            for model in ModelHubUtils.data:
                if (model['language'] in nlu_ref.split(".") or model['language'] in nlp_refrence.split('_')) and \
                        model['name'] == nlp_refrence:
                    return f"https://nlp.johnsnowlabs.com/{model['url']}"
            for model in ModelHubUtils.data:
                # Retry, but no respect to lang
                if model['name'] == nlp_refrence:
                    return f"https://nlp.johnsnowlabs.com/{model['url']}"
        return 'https://nlp.johnsnowlabs.com/models'

    @staticmethod
    def return_json_entry(nlu_ref: str) -> dict:
        """Resolves a Json entry for annlu_refrence.
        :param nlu_ref:  What nlp_refrence to resolve
        :return: Json entry of that nlu reference
        """
        if nlu_ref.split(".")[0] not in nlu.Spellbook.pretrained_models_references.keys():
            nlu_ref = "en." + nlu_ref
        nlp_refrence = ModelHubUtils.NLU_ref_to_NLP_ref(nlu_ref)

        language = nlu_ref.split(".")[0]
        for model in ModelHubUtils.data:
            if model['language'] == language and model["name"] == nlp_refrence:
                return model



#
# # Thin wrapper around sparknlp_jsl classes
# --> Spark NLP annotator Objects
# from johnsnowlabs import Finance, Legal, Medical
# johnsnowlabs.start()
#
# # Under the hood using MedcicalNer/EnterpriseNer Scala class
# fin_ner = Finance.Ner.pretrained().setInputCols() .....
#
# NerDl
# MedicalNerDl
# FinanceNerDl
# # Put in NLU?
# model = MedicalnerDl.pretrained('whaletver')
#
# nlu_pipe = nlu.auto_complete(model)
# nlu_pipe.vanila_pipeline.stages # <----
from sparknlp.common import AnnotatorModel
import sparknlp.annotator
from sparknlp.annotator import *
from sparknlp.internal import ExtendedJavaWrapper

class _FinanceNerModelLoader(ExtendedJavaWrapper):
    def __init__(self, ner_model_path, path, jspark):
        super(_FinanceNerModelLoader, self).__init__(
            "com.johnsnowlabs.nlp.annotators.ner.MedicalNerModel.loadSavedModel", ner_model_path, path, jspark)

class FinanceNer(AnnotatorModel, HasStorageRef, HasBatchedAnnotate):
    name = "MedicalNerModel"
    def __init__(self, classname="com.johnsnowlabs.nlp.annotators.ner.MedicalNerModel", java_model=None):
        super(FinanceNer, self).__init__(
            classname=classname,
            java_model=java_model
        )
        self._setDefault(
            includeConfidence=False,
            includeAllConfidenceScores=False,
            batchSize=8,
            inferenceBatchSize=1
        )

    configProtoBytes = Param(Params._dummy(), "configProtoBytes",
                             "ConfigProto from tensorflow, serialized into byte array. Get with config_proto.SerializeToString()",
                             TypeConverters.toListString)
    includeConfidence = Param(Params._dummy(), "includeConfidence",
                              "whether to include confidence scores in annotation metadata",
                              TypeConverters.toBoolean)
    includeAllConfidenceScores = Param(Params._dummy(), "includeAllConfidenceScores",
                                       "whether to include all confidence scores in annotation metadata or just the score of the predicted tag",
                                       TypeConverters.toBoolean)
    inferenceBatchSize = Param(Params._dummy(), "inferenceBatchSize",
                               "number of sentences to process in a single batch during inference",
                               TypeConverters.toInt)
    classes = Param(Params._dummy(), "classes",
                    "get the tags used to trained this MedicalNerModel",
                    TypeConverters.toListString)

    trainingClassDistribution = Param(Params._dummy(),
                                      "trainingClassDistribution",
                                      "class counts for each of the classes during training",
                                      typeConverter=TypeConverters.identity)

    labelCasing = Param(Params._dummy(), "labelCasing",
                        "Setting all labels of the NER models upper/lower case. values upper|lower",
                        TypeConverters.toString)

    def setConfigProtoBytes(self, b):
        """Sets configProto from tensorflow, serialized into byte array.

        Parameters
        ----------
        b : List[str]
            ConfigProto from tensorflow, serialized into byte array
        """
        return self._set(configProtoBytes=b)

    def setIncludeConfidence(self, value):
        """Sets whether to include confidence scores in annotation metadata, by
        default False.

        Parameters
        ----------
        value : bool
            Whether to include the confidence value in the output.
        """
        return self._set(includeConfidence=value)

    def setIncludeConfidence(self, value):
        """Sets whether to include confidence scores in annotation metadata, by
        default False.

        Parameters
        ----------
        value : bool
            Whether to include the confidence value in the output.
        """
        return self._set(includeConfidence=value)

    def setInferenceBatchSize(self, value):
        """Sets number of sentences to process in a single batch during inference

        Parameters
        ----------
        value : int
           number of sentences to process in a single batch during inference
        """
        return self._set(inferenceBatchSize=value)

    def setLabelCasing(self, value):
        """Setting all labels of the NER models upper/lower case. values upper|lower

        Parameters
        ----------
        value : str
           Setting all labels of the NER models upper/lower case. values upper|lower
        """
        return self._set(labelCasing=value)

    def getTrainingClassDistribution(self):
        return self._call_java('getTrainingClassDistributionJava')

    @staticmethod
    def pretrained(name="ner_dl", lang="en", remote_loc=None):
        from sparknlp.pretrained import ResourceDownloader
        return ResourceDownloader.downloadModel(FinanceNer, name, lang, remote_loc)

    @staticmethod
    def loadSavedModel(ner_model_path, folder, spark_session):
        jModel = _FinanceNerModelLoader(ner_model_path, folder, spark_session._jsparkSession)._java_obj
        return FinanceNer(java_model=jModel)

    @staticmethod
    def pretrained(name="ner_clinical", lang="en", remote_loc="clinical/models"):
        from sparknlp.pretrained import ResourceDownloader
        return ResourceDownloader.downloadModel(FinanceNer, name, lang, remote_loc,
                                                j_dwn='InternalsPythonResourceDownloader')


#from nlu import finance
# finance.FinanceNER


from sparknlp.annotator import E5Embeddings


class E5:
    @staticmethod
    def get_default_model():
        return E5Embeddings.pretrained() \
            .setInputCols(["document"]) \
            .setOutputCol("e5_embeddings")

    @staticmethod
    def get_pretrained_model(name, language, bucket=None):
        return E5Embeddings.pretrained(name,language,bucket) \
            .setInputCols(["document"]) \
            .setOutputCol("e5_embeddings")



"""Helper mappings to resolve storage ref for some healthcare models.
Sometimes Storageref==UID on some Annotators. To resolve to correct embedding provider, we must define a mapping from UID->storageref
"""
mappings = {
'RelationExtractionModel_9c255241fec3':'clinical',
'RelationExtractionModel_14b00157fc1a' : 'clinical',

}



import logging

import nlu
from nlu import Licenses
from nlu.info import AllComponentsInfo
from nlu.spellbook import Spellbook

logger = logging.getLogger('nlu')


def check_if_nlu_ref_is_licensed(nlu_ref):
    """check if a nlu_ref is pointing to a licensed or open source model_anno_obj.
    This works by just checking if the NLU ref points to a healthcare model_anno_obj or not"""
    for lang, universe in Spellbook.healthcare_component_alias_references.items():
        for hc_nlu_ref, hc_nlp_ref in universe.items():
            if hc_nlu_ref == nlu_ref: return True
    for lang, universe in Spellbook.pretrained_healthcare_model_references.items():
        for hc_nlu_ref, hc_nlp_ref in universe.items():
            if hc_nlu_ref == nlu_ref: return True
    return False


def parse_language_from_nlu_ref(nlu_ref):
    """Parse a ISO language identifier from a NLU reference which can be used to load a Spark NLP model_anno_obj"""
    infos = nlu_ref.split('.')
    lang = None
    for split in infos:
        if split in nlu.Spellbook.pretrained_models_references.keys():
            lang = split

    if not lang:
        # translators are internally handled as 'xx'
        if 'translate_to' in nlu_ref and not 't5' in nlu_ref:
            # Special case for translate_to, will be prefixed with 'xx' if not already prefixed with xx because there are only xx.<ISO>.translate_to references
            language = 'xx'
            if nlu_ref[0:3] != 'xx.':
                nlu_reference = 'xx.' + nlu_ref
            logger.info(f'Setting lang as xx for nlu_ref={nlu_reference}')
    if not lang:
        lang = 'en'
    logger.info(f'Parsed Nlu_ref={nlu_ref} as lang={lang}')

    return lang


def nlu_ref_to_nlp_ref(nlu_ref):
    _, _, nlp_ref, _, _ = nlu_ref_to_nlp_metadata(nlu_ref)
    return nlp_ref


def nlu_ref_to_nlp_metadata(nlu_ref, is_recursive_call=False):
    """
    For given NLU ref, returns is_pipe, license_type

    :return: lang, nlu_ref, nlp_ref, license_type, is_pipe
    """
    model_params = None
    lang = parse_language_from_nlu_ref(nlu_ref)
    nlp_ref = None
    license_type = Licenses.open_source
    is_pipe = False
    if 'translate_to' in nlu_ref :
        # We append here xx and set lang as xx  so users don't have to specify it
        lang = 'xx'
        if 'xx' not in nlu_ref:
            nlu_ref = 'xx.' + nlu_ref
    # 1. check if open source pipeline
    if lang in Spellbook.pretrained_pipe_references.keys():
        if nlu_ref in Spellbook.pretrained_pipe_references[lang].keys():
            nlp_ref = Spellbook.pretrained_pipe_references[lang][nlu_ref]
            is_pipe = True
    # 2. check if open source model_anno_obj
    if lang in Spellbook.pretrained_models_references.keys():
        if nlu_ref in Spellbook.pretrained_models_references[lang].keys():
            nlp_ref = Spellbook.pretrained_models_references[lang][nlu_ref]
            logger.info(f'Found Spark NLP reference in pretrained models namespace = {nlp_ref}')

    # 3. check if open source alias
    if nlu_ref in Spellbook.component_alias_references.keys():
        sparknlp_data = Spellbook.component_alias_references[nlu_ref]
        nlp_ref = sparknlp_data[0]
        is_pipe = 'component_list' in sparknlp_data[1]
        if len(sparknlp_data) == 3:
            model_params = sparknlp_data[2]
    # 4. check if healthcare pipe
    if lang in Spellbook.pretrained_healthcare_pipe_references.keys():
        if nlu_ref in Spellbook.pretrained_healthcare_pipe_references[lang].keys():
            nlp_ref = Spellbook.pretrained_healthcare_pipe_references[lang][nlu_ref]
            license_type = Licenses.hc
            is_pipe = True

    # 5. check if healthcare model_anno_obj
    if lang in Spellbook.pretrained_healthcare_model_references.keys():
        if nlu_ref in Spellbook.pretrained_healthcare_model_references[lang].keys():
            nlp_ref = Spellbook.pretrained_healthcare_model_references[lang][nlu_ref]
            license_type = Licenses.hc

    # 6. check if healthcare alias
    if nlu_ref in Spellbook.healthcare_component_alias_references.keys():
        sparknlp_data = Spellbook.healthcare_component_alias_references[nlu_ref]
        nlp_ref = sparknlp_data[0]
        is_pipe = 'component_list' in sparknlp_data[1]
        license_type = Licenses.hc

    # 7. check if ocr model_anno_obj
    if nlu_ref in Spellbook.ocr_model_references.keys():
        nlp_ref = Spellbook.ocr_model_references[nlu_ref]
        license_type = Licenses.ocr

    # Check if multi lingual ner
    if not nlp_ref and 'ner' in nlu_ref:
        all_component_info = AllComponentsInfo()
        if lang in all_component_info.all_multi_lang_base_ner_languages:
            lang = 'xx'
            nlp_ref = 'ner_wikiner_glove_840B_300'
            nlu_ref = 'xx.ner.wikiner_glove_840B_300'
        if lang in all_component_info.all_multi_lang_xtreme_ner_languages:
            lang = 'xx'
            nlp_ref = 'ner_xtreme_glove_840B_300'
            nlu_ref = 'xx.ner.xtreme_glove_840B_300'

    # Search again but with en. prefixed, enables all refs to work without en prefix
    if not nlp_ref and not is_recursive_call:
        return nlu_ref_to_nlp_metadata('en.' + nlu_ref, is_recursive_call=True)
    #
    # if is_licensed:
    #     if not auth_utils.is_authorized_environment() and is_licensed:
    #         print(f"The nlu_ref=[{nlu_ref}] is pointing to a licensed Spark NLP Annotator or Model [{nlp_ref}]. \n"
    #               f"Your environment does not seem to be Authorized!\n"
    #               f"Please RESTART your Python environment and run nlu.auth(SPARK_NLP_LICENSE,AWS_ACCESS_KEY_ID,AWS_SECRET_ACCESS_KEY,JSL_SECRET)  \n"
    #               f"with your corrosponding credentials. If you have no credentials you can get a trial version with credentials here https://www.johnsnowlabs.com/spark-nlp-try-free/ \n"
    #               f"Or contact us at contact@jonsnowlabs.com\n"
    #               f"NLU will ignore this error and continue running, but you will encounter errors most likely. ")

    return lang, nlu_ref, nlp_ref, license_type, is_pipe, model_params



from sparknlp.annotator import DistilBertForQuestionAnswering


class SpanDistilBertClassifier:
    @staticmethod
    def get_default_model():
        return DistilBertForQuestionAnswering.pretrained() \
            .setInputCols(["document_question", "context"]) \
            .setOutputCol("answer")

    @staticmethod
    def get_pretrained_model(name, language, bucket=None):
        return DistilBertForQuestionAnswering.pretrained(name, language, bucket) \
            .setInputCols(["document_question", "context"]) \
            .setOutputCol("answer")



import importlib
import json
import site
from importlib import reload

from nlu.utils.environment.env_utils import *


def ocr_version_is_missmatch(secret_version):
    """Check if installed OCR version is the same as the OCR secret
    :param secret_version: ocr secret version
    :return: True if missmatch, False if versions are fien """
    try:
        import sparkocr
    except ImportError:
        # Import failed, there can be no version missmatch
        return False
    # check versions
    installed_version = sparkocr.version()
    if installed_version == secret_version:
        return False
    return True


def healthcare_version_is_missmatch(secret_version):
    """Check if installed Spark NLP healthcare version is the same as the Healthcare secret
    :param secret_version: healthcare secret version
    :return: True if missmatch, False if versions are fine"""
    try:
        import sparknlp_jsl
    except ImportError:
        # Import failed, there can be no version missmatch
        return False
    # check versions
    installed_version = sparknlp_jsl.version()
    if installed_version == secret_version:
        return False
    return True


def uninstall_lib(pip_package_name):
    cmd = f'{sys.executable} -m pip uninstall {pip_package_name} -y '
    os.system(cmd)
    reload(site)


def check_if_secret_missmatch_and_uninstall_if_bad(secret_version, module_name, package_name):
    """Check if OCR/Healthcare lib installed version match up with the secrets provided.
    If not, this will uninstall the missmaching library
    :param module_name: module import name
    :param package_name: pipe package name
    :param secret_version: healthcare/ocr secret version provided
    :return: True if missmatch was uninstalled, False if no missmatch found and nothing was done
    """


    try:
        importlib.import_module(module_name)
    except ImportError:
        # Import failed, there can be no version missmatch
        return False
    if module_name == 'sparknlp_jsl':
        # get versions
        import sparknlp_jsl
        installed_version = sparknlp_jsl.version()
        if installed_version == secret_version:
            return False
    elif module_name == 'sparkocr':
        # get versions
        import sparkocr
        installed_version = sparkocr.version()
        if installed_version == secret_version:
            return False
    else:
        raise ValueError(f'Invalid module_name=={module_name}')

    print(
        f"Installed {module_name}=={installed_version} version not matching provided secret version=={secret_version}. "
        f"Uninstalling it..")
    # version missmatch, uninstall shell
    uninstall_lib(package_name)
    return True


def import_or_install_licensed_lib(JSL_SECRET, lib='healthcare'):
    """ Install Spark-NLP-Healthcare PyPI Package in current environment if it cannot be imported and license
    provided """
    hc_module_name = 'sparknlp_jsl'
    hc_pip_package_name = 'spark-nlp-jsl'
    hc_display_name = ' Spark NLP for Healthcare'
    ocr_pip_package_name = 'spark-ocr'
    ocr_module_name = 'sparkocr'
    ocr_display_name = ' Spark OCR'
    get_deps = True
    lib_version = JSL_SECRET.split('-')[0]
    missmatch = False

    if lib == 'healthcare':
        target_import = hc_module_name
        target_install = hc_pip_package_name
        display_name = hc_display_name
        missmatch = check_if_secret_missmatch_and_uninstall_if_bad(lib_version, hc_module_name, hc_pip_package_name)
    elif lib == 'ocr':
        target_import = ocr_module_name
        target_install = ocr_pip_package_name
        display_name = ocr_display_name
        # OCR version is suffixed with spark version
        if is_env_pyspark_2_3():
            lib_version = lib_version + '+spark23'
        if is_env_pyspark_2_4():
            lib_version = lib_version + '+spark24'
        if is_env_pyspark_3_0() or is_env_pyspark_3_1():
            lib_version = lib_version + '+spark30'
        # get_deps = False

    else:
        raise ValueError(f'Invalid install licensed install target ={lib}')

    try:
        # Try importing, if it fails install the pacakge
        importlib.import_module(target_import)
        if missmatch:
            raise ImportError
    except ImportError:
        # Install package since its missing
        import pip
        print(f"{display_name} could not be imported. Installing latest {target_install} PyPI package via pip...")
        import pyspark
        pip_major_version = int(pip.__version__.split('.')[0])
        if pip_major_version in [10, 18, 19, 20]:
            # for these versions pip module does not support installing from Python, we install via OS command.
            cmd = f'{sys.executable} -m pip install {target_install}=={lib_version} --extra-index-url https://pypi.johnsnowlabs.com/{JSL_SECRET}'
            if not get_deps:
                cmd = cmd + '--no-deps'
            os.system(cmd)
        else:
            params = ['install', f'{target_install}=={lib_version}', '--extra-index-url',
                      f'https://pypi.johnsnowlabs.com/{JSL_SECRET}']
            if not get_deps:
                params.append('--no-deps')
            pip.main(params)
    finally:
        # Import module after installing package

        reload(site)
        globals()[target_import] = importlib.import_module(target_import)


def authenticate_enviroment_HC(SPARK_NLP_LICENSE, AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY):
    """Set Secret environ variables for Spark Context"""
    os.environ['SPARK_NLP_LICENSE'] = SPARK_NLP_LICENSE
    os.environ['AWS_ACCESS_KEY_ID'] = AWS_ACCESS_KEY_ID
    os.environ['AWS_SECRET_ACCESS_KEY'] = AWS_SECRET_ACCESS_KEY


def authenticate_enviroment_OCR(SPARK_OCR_LICENSE, AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY):
    """Set Secret environ variables for Spark Context"""
    os.environ['SPARK_OCR_LICENSE'] = SPARK_OCR_LICENSE
    os.environ['AWS_ACCESS_KEY_ID'] = AWS_ACCESS_KEY_ID
    os.environ['AWS_SECRET_ACCESS_KEY'] = AWS_SECRET_ACCESS_KEY


def authenticate_enviroment_HC_and_OCR(SPARK_NLP_LICENSE, SPARK_OCR_LICENSE, AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY):
    """Set Secret environ variables for Spark Context"""
    authenticate_enviroment_HC(SPARK_NLP_LICENSE, AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY)
    authenticate_enviroment_OCR(SPARK_OCR_LICENSE, AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY)


def get_authenticated_spark_HC(HC_LICENSE, HC_SECRET, AWS_ACCESS_KEY, AWS_SECRET_KEY, gpu):
    import_or_install_licensed_lib(HC_SECRET, 'healthcare')
    authenticate_enviroment_HC(HC_LICENSE, AWS_ACCESS_KEY, AWS_SECRET_KEY)
    import sparknlp
    import sparknlp_jsl
    params = {"spark.driver.memory": "16G",
              "spark.kryoserializer.buffer.max": "2000M",
              "spark.driver.maxResultSize": "2000M"}

    if is_env_pyspark_3_x():
        return sparknlp_jsl.start(HC_SECRET, gpu=gpu, public=sparknlp.version(), params=params)
    else:
        raise ValueError(f"Current Spark version {get_pyspark_version()} not supported!")


def get_authenticated_spark_OCR(OCR_LICENSE, OCR_SECRET, AWS_ACCESS_KEY, AWS_SECRET_KEY, gpu):
    import_or_install_licensed_lib(OCR_SECRET, 'ocr')
    authenticate_enviroment_OCR(OCR_LICENSE, AWS_ACCESS_KEY, AWS_SECRET_KEY)
    import sparkocr
    import sparknlp
    params = {"spark.driver.memory": "16G", "spark.kryoserializer.buffer.max": "2000M",
              "spark.driver.maxResultSize": "2000M"}
    OS_version = sparknlp.version()
    spark = sparkocr.start(secret=OCR_SECRET, nlp_version=OS_version, )
    spark.sparkContext.setLogLevel('ERROR')


def get_authenticated_spark_HC_and_OCR(HC_LICENSE, HC_SECRET, OCR_LICENSE, OCR_SECRET, AWS_ACCESS_KEY, AWS_SECRET_KEY,
                                       gpu):
    import_or_install_licensed_lib(HC_SECRET, 'healthcare')
    import_or_install_licensed_lib(OCR_SECRET, 'ocr')
    authenticate_enviroment_HC_and_OCR(HC_LICENSE, OCR_LICENSE, AWS_ACCESS_KEY, AWS_SECRET_KEY)
    import sparkocr
    import sparknlp
    params = {"spark.driver.memory": "16G", "spark.kryoserializer.buffer.max": "2000M",
              "spark.driver.maxResultSize": "2000M"}

    HC_version = HC_SECRET.split('-')[0]
    OS_version = sparknlp.version()
    spark = sparkocr.start(secret=OCR_SECRET, nlp_secret=HC_SECRET, nlp_version=OS_version, nlp_internal=HC_version)
    spark.sparkContext.setLogLevel('ERROR')


def get_authenticated_spark(SPARK_NLP_LICENSE, AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY, JSL_SECRET, gpu=False, ):
    """
    Authenticates environment if not already done so and returns Spark Context with Healthcare Jar loaded
    0. If no Spark-NLP-Healthcare, install it via PyPi
    1. If not auth, run authenticate_enviroment()

    """
    import sparknlp
    authenticate_enviroment_HC(SPARK_NLP_LICENSE, AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY)
    import_or_install_licensed_lib(JSL_SECRET)
    import sparknlp_jsl
    params = {"spark.driver.memory": "16G",
              "spark.kryoserializer.buffer.max": "2000M",
              "spark.driver.maxResultSize": "2000M"}
    if is_env_pyspark_3_x():
        return sparknlp_jsl.start(JSL_SECRET, gpu=gpu, public=sparknlp.version(), params=params)
    else:
        raise ValueError(f"Current Spark version {get_pyspark_version()} not supported!")


def is_authorized_environment():
    """Check if auth secrets are set in environment"""
    SPARK_NLP_LICENSE = os.getenv('SPARK_NLP_LICENSE')
    AWS_ACCESS_KEY_ID = os.getenv('AWS_ACCESS_KEY_ID')
    AWS_SECRET_ACCESS_KEY = os.getenv('AWS_SECRET_ACCESS_KEY')
    return None not in [SPARK_NLP_LICENSE, AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY]





def auth(HEALTHCARE_LICENSE_OR_JSON_PATH='/content/spark_nlp_for_healthcare.json', AWS_ACCESS_KEY_ID='',
         AWS_SECRET_ACCESS_KEY='', HEALTHCARE_SECRET='', OCR_LICENSE='', OCR_SECRET='', gpu=False):
    """ Authenticate environment for JSL Licensed models
    Installs NLP-Healthcare if not in environment detected
    Either provide path to spark_nlp_for_healthcare.json file as first param or manually enter them,
    HEALTHCARE_LICENSE_OR_JSON_PATH,AWS_ACCESS_KEY_ID,AWS_SECRET_ACCESS_KEY,HEALTHCARE_SECRET .
    Set gpu=true if you want to enable GPU mode
    """

    def has_empty_strings(iterable):
        """Check for a given list of strings, whether it has any empty strings or not"""
        return all(x == '' for x in iterable)

    hc_creds = [HEALTHCARE_LICENSE_OR_JSON_PATH, HEALTHCARE_SECRET]
    ocr_creds = [OCR_LICENSE, OCR_SECRET]
    aws_creds = [AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY]

    if os.path.exists(HEALTHCARE_LICENSE_OR_JSON_PATH):
        # Credentials provided via JSON file
        with open(HEALTHCARE_LICENSE_OR_JSON_PATH, encoding="utf8") as json_file:
            j = json.load(json_file)
            if 'SPARK_NLP_LICENSE' in j.keys() and 'SPARK_OCR_LICENSE' in j.keys():
                # HC and OCR creds provided
                get_authenticated_spark_HC_and_OCR(j['SPARK_NLP_LICENSE'], j['SECRET'],
                                                              j['SPARK_OCR_LICENSE'], j['SPARK_OCR_SECRET'],
                                                              j['AWS_ACCESS_KEY_ID'], j['AWS_SECRET_ACCESS_KEY'], gpu)

                return True

            if 'SPARK_NLP_LICENSE' in j.keys() and 'SPARK_OCR_LICENSE' not in j.keys():
                # HC creds provided but no OCR
                get_authenticated_spark_HC(j['SPARK_NLP_LICENSE'], j['SECRET'], j['AWS_ACCESS_KEY_ID'],
                                                      j['AWS_SECRET_ACCESS_KEY'], gpu)
                return True

            if 'SPARK_NLP_LICENSE' not in j.keys() and 'SPARK_OCR_LICENSE' in j.keys():
                # OCR creds provided but no HC
                get_authenticated_spark_OCR(j['SPARK_OCR_LICENSE'], j['SPARK_OCR_SECRET'],
                                                       j['AWS_ACCESS_KEY_ID'], j['AWS_SECRET_ACCESS_KEY'], gpu)
                return True

            get_authenticated_spark(gpu)
        return True
    else:
        # Credentials provided as parameter
        if not has_empty_strings(hc_creds) and not has_empty_strings(ocr_creds) and not has_empty_strings(aws_creds):
            # HC + OCR credentials provided
            get_authenticated_spark_HC_and_OCR(HEALTHCARE_LICENSE_OR_JSON_PATH, HEALTHCARE_SECRET,
                                                          OCR_LICENSE,
                                                          OCR_SECRET, AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY, gpu)
            return True

        elif not has_empty_strings(hc_creds) and has_empty_strings(ocr_creds) and not has_empty_strings(aws_creds):
            # HC creds provided, but no HC
            get_authenticated_spark_HC(HEALTHCARE_LICENSE_OR_JSON_PATH, HEALTHCARE_SECRET, AWS_ACCESS_KEY_ID,
                                                  AWS_SECRET_ACCESS_KEY, gpu)
            return True
        elif has_empty_strings(hc_creds) and not has_empty_strings(ocr_creds) and not has_empty_strings(aws_creds):
            # OCR creds provided but no HC
            get_authenticated_spark_OCR(OCR_LICENSE, OCR_SECRET, AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY,
                                                   gpu)
            return True

    return False



from sparknlp_display import NerVisualizer, DependencyParserVisualizer
from sparknlp.annotator import NerConverter, DependencyParserModel, TypedDependencyParserModel, PerceptronModel
from sparknlp.base import DocumentAssembler
from nlu.universe.feature_node_ids import NLP_NODE_IDS

class VizUtilsOS():
    """Utils for interfacing with the Spark-NLP-Display lib and vizzing Open Source Components - Open source"""
    HTML_WRAPPER = """<div class="scroll entities" style="overflow-x: auto; border: 1px solid #e6e9ef; border-radius: 0.25rem; padding: 1rem; margin-bottom: 2.5rem; white-space:pre-wrap">{}</div>"""

    @staticmethod
    def infer_viz_open_source(pipe) -> str:
        """For a given NLUPipeline with only open source components, infers which visualizations are applicable. """
        for c in pipe.components:
            if isinstance(c.model, NerConverter): return 'ner'
            if isinstance(c.model, DependencyParserModel): return 'dep'

    @staticmethod
    def viz_ner(anno_res, pipe, labels=None, viz_colors={}, is_databricks_env=False, write_to_streamlit=False,
                streamlit_key='RANDOM',  user_ner_col=None):
        """Infer columns required for ner viz and then viz it.
        viz_colors :  set label colors by specifying hex codes , i.e. viz_colors =  {'LOC':'#800080', 'PER':'#77b5fe'}
        labels : only allow these labels to be displayed. (default: [] - all labels will be displayed)
        """
        document_col, entities_col = VizUtilsOS.infer_ner_dependencies(pipe)
        ner_vis = NerVisualizer()
        ner_vis.set_label_colors(viz_colors)
        if user_ner_col :
            entities_col = user_ner_col
        if write_to_streamlit:
            import streamlit as st
            HTML = ner_vis.display(anno_res, label_col=entities_col, document_col=document_col, labels=labels,
                                   return_html=True)
            CSS, HTML = HTML.split('</style>')
            CSS = CSS + '</style>'
            HTML = f'<div> {HTML} '
            st.markdown(CSS, unsafe_allow_html=True)
            st.markdown(VizUtilsOS.HTML_WRAPPER.format(HTML), unsafe_allow_html=True)

        elif not is_databricks_env:
            ner_vis.display(anno_res, label_col=entities_col, document_col=document_col, labels=labels)
        else:
            return ner_vis.display(anno_res, label_col=entities_col, document_col=document_col, labels=labels,
                                   return_html=True)

    @staticmethod
    def infer_ner_dependencies(pipe):
        """Finds entities and doc cols for ner viz"""
        # TODO FIX
        doc_component = None
        entities_component = None
        for c in pipe.components:
            if isinstance(c.model, NerConverter):        entities_component = c
            if isinstance(c.model, DocumentAssembler):   doc_component = c

        document_col = doc_component.spark_output_column_names[0]
        entities_col = entities_component.spark_output_column_names[0]
        return document_col, entities_col

    @staticmethod
    def viz_dep(anno_res, pipe, is_databricks_env, write_to_streamlit, streamlit_key='RANDOM',
                user_pos_col=None, user_dep_untyped_col=None, user_dep_typed_col=None
                ):
        """Viz dep result"""
        pos_col, dep_typ_col, dep_untyp_col = VizUtilsOS.infer_dep_dependencies(pipe)
        dependency_vis = DependencyParserVisualizer()
        if user_pos_col :
            pos_col = user_pos_col

        if user_dep_typed_col:
            dep_typ_col = user_dep_typed_col

        if user_dep_untyped_col:
            dep_untyp_col = dep_untyp_col

        if write_to_streamlit:
            import streamlit as st
            SVG = dependency_vis.display(anno_res, pos_col=pos_col, dependency_col=dep_untyp_col,
                                         dependency_type_col=dep_typ_col, return_html=True)
            # st.markdown(SVG, unsafe_allow_html=True)
            st.markdown(VizUtilsOS.HTML_WRAPPER.format(SVG), unsafe_allow_html=True)

        elif not is_databricks_env:
            dependency_vis.display(anno_res, pos_col=pos_col, dependency_col=dep_untyp_col,
                                   dependency_type_col=dep_typ_col)
        else:
            return dependency_vis.display(anno_res, pos_col=pos_col, dependency_col=dep_untyp_col,
                                          dependency_type_col=dep_typ_col, return_html=True)

    @staticmethod
    def infer_dep_dependencies(pipe):
        """Finds entities,pos,dep_typed,dep_untyped and  doc cols for dep viz viz"""
        # doc_component      = None
        pos_component = None
        dep_untyped_component = None
        dep_typed_component = None
        for c in pipe.components:
            if c.name == NLP_NODE_IDS.POS :
                pos_component = c
            if c.name == NLP_NODE_IDS.TYPED_DEPENDENCY_PARSER :
                dep_typed_component = c
            if c.name == NLP_NODE_IDS.UNTYPED_DEPENDENCY_PARSER :
                dep_untyped_component = c

        pos_col = pos_component.spark_output_column_names[0]
        dep_typ_col = dep_typed_component.spark_output_column_names[0]
        dep_untyp_col = dep_untyped_component.spark_output_column_names[0]
        return pos_col, dep_typ_col, dep_untyp_col



"""
Collection JSL Annotator IDs used by NLU internally
"""
from nlu.universe.atoms import JslAnnoId


class NLP_NODE_IDS:
    """All avaiable Feature nodes in NLP..
    Used to cast the pipeline dependency resolution algorithm into an abstract graph
    """
    # Visual Document Understanding

    BIG_TEXT_MATCHER = JslAnnoId('big_text_matcher')
    CHUNK2DOC = JslAnnoId('chunk2doc')
    CHUNK_EMBEDDINGS_CONVERTER = JslAnnoId('chunk_embeddings_converter')
    CHUNK_TOKENIZER = JslAnnoId('chunk_tokenizer')
    CHUNKER = JslAnnoId('chunker')
    CLASSIFIER_DL = JslAnnoId('classifier_dl')
    CONTEXT_SPELL_CHECKER = JslAnnoId('context_spell_checker')
    DATE_MATCHER = JslAnnoId('date_matcher')
    UNTYPED_DEPENDENCY_PARSER = JslAnnoId('untyped_dependency_parser')
    TYPED_DEPENDENCY_PARSER = JslAnnoId('typed_dependency_parser')
    DOC2CHUNK = JslAnnoId('doc2chunk')
    DOC2VEC = JslAnnoId('doc2vec')
    TRAIANBLE_DOC2VEC = JslAnnoId('trainable_doc2vec')
    DOCUMENT_ASSEMBLER = JslAnnoId('document_assembler')
    AUDIO_ASSEMBLER = JslAnnoId('audio_assembler')
    WAV2VEC_FOR_CTC = JslAnnoId('wav2vec_for_ctc')
    HUBERT_FOR_CTC = JslAnnoId('hubert_for_ctc')
    WHISPER_FOR_CTC = JslAnnoId('whisper_for_ctc')
    TABLE_ASSEMBLER = JslAnnoId('table_assembler')
    TAPAS_FOR_QA = JslAnnoId('tapas')
    MULTI_DOCUMENT_ASSEMBLER = JslAnnoId('multi_document_assembler')
    IMAGE_ASSEMBLER = JslAnnoId("image_assembler")

    ALBERT_FOR_QUESTION_ANSWERING = JslAnnoId('albert_for_question_answering')
    CAMEMBERT_FOR_QUESTION_ANSWERING = JslAnnoId('camembert_for_question_answering')
    BERT_FOR_QUESTION_ANSWERING = JslAnnoId('bert_for_question_answering')
    DE_BERTA_FOR_QUESTION_ANSWERING = JslAnnoId('de_berta_for_question_answering')
    DISTIL_BERT_FOR_QUESTION_ANSWERING = JslAnnoId('distil_bert_for_question_answering')
    LONGFORMER_FOR_QUESTION_ANSWERING = JslAnnoId('longformer_for_question_answering')
    ROBERTA_FOR_QUESTION_ANSWERING = JslAnnoId('roberta_for_question_answering')
    XLM_ROBERTA_FOR_QUESTION_ANSWERING = JslAnnoId('xlm_roberta_for_question_answering')
    SPAN_BERT_COREF_MODEL = JslAnnoId('span_bert_coref_model')



    DOCUMENT_NORMALIZER = JslAnnoId('document_normalizer')
    EMBEDDINGS_FINISHER = JslAnnoId('embeddings_finisher')
    ENTITY_RULER = JslAnnoId('entitiy_ruler')
    FINISHER = JslAnnoId('FINISHER')
    GRAPH_EXTRACTION = JslAnnoId('graph_extraction')
    GRAPH_FINISHER = JslAnnoId('graph_finisher')
    LANGUAGE_DETECTOR_DL = JslAnnoId('language_detector_dl')
    LEMMATIZER = JslAnnoId('lemmatizer')
    MULTI_CLASSIFIER_DL = JslAnnoId('multi_classifier_dl')
    MULTI_DATE_MATCHER = JslAnnoId('multi_date_matcher')
    N_GRAMM_GENERATOR = JslAnnoId('n_gramm_generator')
    NER_CONVERTER = JslAnnoId('ner_converter')
    NER_CRF = JslAnnoId('ner_crf')
    NER_DL = JslAnnoId('ner_dl')
    NER_OVERWRITER = JslAnnoId('ner_overwriter')
    NORMALIZER = JslAnnoId('normalizer')
    NORVIG_SPELL_CHECKER = JslAnnoId('norvig_spell_checker')
    POS = JslAnnoId('pos')
    RECURISVE_TOKENIZER = JslAnnoId('recursive_tokenizer')
    REGEX_MATCHER = JslAnnoId('regex_matcher')
    REGEX_TOKENIZER = JslAnnoId('regex_tokenizer')

    SENTENCE_DETECTOR = JslAnnoId('sentence_detector')
    SENTENCE_DETECTOR_DL = JslAnnoId('sentence_detector_dl')
    SENTENCE_EMBEDDINGS_CONVERTER = JslAnnoId('sentence_embeddings_converter')
    STEMMER = JslAnnoId('stemmer')
    STOP_WORDS_CLEANER = JslAnnoId('stop_words_cleaner')
    SYMMETRIC_DELETE_SPELLCHECKER = JslAnnoId('symmetric_delete_spellchecker')
    TEXT_MATCHER = JslAnnoId('text_matcher')
    TOKEN2CHUNK = JslAnnoId('token2chunk')
    TOKEN_ASSEMBLER = JslAnnoId('token_assembler')
    TOKENIZER = JslAnnoId('tokenizer')
    SENTIMENT_DL = JslAnnoId('sentiment_dl')
    SENTIMENT_DETECTOR = JslAnnoId('sentiment_detector')
    VIVEKN_SENTIMENT = JslAnnoId('vivekn_sentiment')
    WORD_EMBEDDINGS = JslAnnoId('word_embeddings')
    WORD_SEGMENTER = JslAnnoId('word_segmenter')
    YAKE_KEYWORD_EXTRACTION = JslAnnoId('yake_keyword_extraction')
    ALBERT_EMBEDDINGS = JslAnnoId('albert_embeddings')
    ALBERT_FOR_TOKEN_CLASSIFICATION = JslAnnoId('albert_for_token_classification')
    BERT_EMBEDDINGS = JslAnnoId('bert_embeddings')
    CAMEMBERT_EMBEDDINGS = JslAnnoId('camenbert_embeddings')
    BERT_FOR_TOKEN_CLASSIFICATION = JslAnnoId('bert_for_token_classification')
    CAMEMBERT_FOR_TOKEN_CLASSIFICATION = JslAnnoId('camenbert_for_token_classification')
    CAMEMBERT_FOR_SEQUENCE_CLASSIFICATION = JslAnnoId('camenbert_for_sequence_classification')
    E5_SENTENCE_EMBEDDINGS = JslAnnoId('e5_sentence_embeddings')
    BERT_SENTENCE_EMBEDDINGS = JslAnnoId('bert_sentence_embeddings')
    DISTIL_BERT_EMBEDDINGS = JslAnnoId('distil_bert_embeddings')
    DISTIL_BERT_FOR_TOKEN_CLASSIFICATION = JslAnnoId('distil_bert_for_token_classification')
    DISTIL_BERT_FOR_SEQUENCE_CLASSIFICATION = JslAnnoId('distil_bert_for_sequence_classification')
    BERT_FOR_SEQUENCE_CLASSIFICATION = JslAnnoId('bert_for_sequence_classification')
    ELMO_EMBEDDINGS = JslAnnoId('elmo_embeddings')
    LONGFORMER_EMBEDDINGS = JslAnnoId('longformer_embeddings')
    LONGFORMER_FOR_TOKEN_CLASSIFICATION = JslAnnoId('longformer_for_token_classification')
    MARIAN_TRANSFORMER = JslAnnoId('marian_transformer')
    ROBERTA_EMBEDDINGS = JslAnnoId('roberta_embeddings')
    ROBERTA_FOR_TOKEN_CLASSIFICATION = JslAnnoId('roberta_for_token_classification')
    ROBERTA_SENTENCE_EMBEDDINGS = JslAnnoId('roberta_sentence_embeddings')
    T5_TRANSFORMER = JslAnnoId('t5_transformer')
    VIT_IMAGE_CLASSIFICATION = JslAnnoId("vit_image_classification")
    CONVNEXT_IMAGE_CLASSIFICATION = JslAnnoId("convnext_image_classification")
    SWIN_IMAGE_CLASSIFICATION = JslAnnoId("swin_image_classification")
    BART_TRANSFORMER = JslAnnoId("bart_transformer")
    INSTRUCTOR_SENTENCE_EMBEDDINGS = JslAnnoId('instructor_sentence_embeddings')

    MPNET_SENTENCE_EMBEDDINGS = JslAnnoId('mpnet_sentence_embeddings')





    DISTIL_BERT_FOR_ZERO_SHOT_CLASSIFICATION = JslAnnoId('distil_bert_zero_shot')
    BERT_FOR_ZERO_SHOT_CLASSIFICATION = JslAnnoId('bert_zero_shot')
    ROBERTA_FOR_ZERO_SHOT_CLASSIFICATION = JslAnnoId('roberta_zero_shot')
    UNIVERSAL_SENTENCE_ENCODER = JslAnnoId('universal_sentence_encoder')
    XLM_ROBERTA_EMBEDDINGS = JslAnnoId('xlm_roberta_embeddings')
    XLM_ROBERTA_FOR_TOKEN_CLASSIFICATION = JslAnnoId('xlm_roberta_for_token_classification')
    XLM_ROBERTA_SENTENCE_EMBEDDINGS = JslAnnoId('xlm_roberta_sentence_embeddings')
    XLNET_EMBEDDINGS = JslAnnoId('xlnet_embeddings')
    XLNET_FOR_TOKEN_CLASSIFICATION = JslAnnoId('xlnet_for_token_classification')
    XLM_ROBERTA_FOR_SEQUENCE_CLASSIFICATION = JslAnnoId('xlm_roberta_for_sequence_classification')
    ROBERTA_FOR_SEQUENCE_CLASSIFICATION = JslAnnoId('roberta_for_sequence_classification')
    LONGFORMER_FOR_SEQUENCE_CLASSIFICATION = JslAnnoId('longformer_for_sequence_classification')
    ALBERT_FOR_SEQUENCE_CLASSIFICATION = JslAnnoId('albert_for_sequence_classification')
    XLNET_FOR_SEQUENCE_CLASSIFICATION = JslAnnoId('xlnet_for_sequence_classification')
    GPT2 = JslAnnoId('gpt2')
    WORD_2_VEC = JslAnnoId('word_2_vec')
    DEBERTA_WORD_EMBEDDINGS = JslAnnoId('deberta')
    DEBERTA_FOR_SEQUENCE_CLASSIFICATION = JslAnnoId('deberta_for_sequence_classification')
    DEBERTA_FOR_TOKEN_CLASSIFICATION = JslAnnoId('deberta_for_token_classification')
    COREF_SPAN_BERT = JslAnnoId('coref_spanbert')


    TRAINABLE_CONTEXT_SPELL_CHECKER = JslAnnoId('trainable_context_spell_checker')
    TRAINABLE_VIVEKN_SENTIMENT = JslAnnoId('trainable_vivekn_sentiment')
    TRAINABLE_SENTIMENT_DL = JslAnnoId('trainable_sentiment_dl')
    TRAINABLE_CLASSIFIER_DL = JslAnnoId('trainable_classifier_dl')
    TRAINABLE_MULTI_CLASSIFIER_DL = JslAnnoId('trainable_multi_classifier_dl')
    TRAINABLE_NER_DL = JslAnnoId('trainable_ner_dl')
    TRAINABLE_NER_CRF = JslAnnoId('trainable_ner_crf')
    TRAINABLE_POS = JslAnnoId('trainable_pos')
    TRAINABLE_DEP_PARSE_TYPED = JslAnnoId('trainable_dependency_parser')
    TRAINABLE_DEP_PARSE_UN_TYPED = JslAnnoId('trainable_dependency_parser_untyped')
    TRAINABLE_DOC2VEC = JslAnnoId('trainable_doc2vec')
    TRAINABLE_ENTITY_RULER = JslAnnoId('trainable_entity_ruler')
    TRAINABLE_LEMMATIZER = JslAnnoId('trainable_lemmatizer')
    TRAINABLE_NORMALIZER = JslAnnoId('trainable_normalizer')
    TRAINABLE_NORVIG_SPELL_CHECKER = JslAnnoId('trainable_norvig_spell')
    TRAINABLE_RECURISVE_TOKENIZER = JslAnnoId('trainable_recursive_tokenizer')
    TRAINABLE_REGEX_MATCHER = JslAnnoId('trainable_regex_tokenizer')
    TRAINABLE_SENTENCE_DETECTOR_DL = JslAnnoId('trainable_sentence_detector_dl')
    TRAINABLE_SENTIMENT = JslAnnoId('trainable_sentiment')
    TRAINABLE_WORD_EMBEDDINGS = JslAnnoId('trainable_word_embeddings')
    TRAINABLE_SYMMETRIC_DELETE_SPELLCHECKER = JslAnnoId('trainable_symmetric_spell_checker')
    TRAINABLE_TEXT_MATCHER = JslAnnoId('trainable_text_matcher')
    TRAINABLE_TOKENIZER = JslAnnoId('trainable_tokenizer')
    TRAINABLE_WORD_SEGMENTER = JslAnnoId('trainable_word_segmenter')
    BERT_SENTENCE_CHUNK_EMBEDDINGS = JslAnnoId('bert_sentence_chunk_embeddings')


    PARTIALLY_IMPLEMENTED = JslAnnoId('partially_implemented')
    PARTIAL_AssertionFilterer = JslAnnoId('PARTIAL_AssertionFilterer')
    PARTIAL_ChunkConverter = JslAnnoId('PARTIAL_ChunkConverter')
    PARTIAL_ChunkKeyPhraseExtraction = JslAnnoId('PARTIAL_ChunkKeyPhraseExtraction')
    PARTIAL_ChunkSentenceSplitter = JslAnnoId('PARTIAL_ChunkSentenceSplitter')
    PARTIAL_ChunkFiltererApproach = JslAnnoId('PARTIAL_ChunkFiltererApproach')
    PARTIAL_ChunkFiltererApproach = JslAnnoId('PARTIAL_ChunkFiltererApproach')
    PARTIAL_ChunkFilterer = JslAnnoId('PARTIAL_ChunkFilterer')
    PARTIAL_ChunkMapperApproach = JslAnnoId('PARTIAL_ChunkMapperApproach')
    PARTIAL_ChunkMapperApproach = JslAnnoId('PARTIAL_ChunkMapperApproach')
    PARTIAL_ChunkMapperFilterer = JslAnnoId('PARTIAL_ChunkMapperFilterer')
    PARTIAL_DocumentLogRegClassifierApproach = JslAnnoId('PARTIAL_DocumentLogRegClassifierApproach')
    PARTIAL_DocumentLogRegClassifierApproach = JslAnnoId('PARTIAL_DocumentLogRegClassifierApproach')
    PARTIAL_DocumentLogRegClassifierModel = JslAnnoId('PARTIAL_DocumentLogRegClassifierModel')
    PARTIAL_ContextualParserApproach = JslAnnoId('PARTIAL_ContextualParserApproach')
    PARTIAL_ContextualParserApproach = JslAnnoId('PARTIAL_ContextualParserApproach')
    PARTIAL_ReIdentification = JslAnnoId('PARTIAL_ReIdentification')
    PARTIAL_NerDisambiguator = JslAnnoId('PARTIAL_NerDisambiguator')
    PARTIAL_NerDisambiguatorModel = JslAnnoId('PARTIAL_NerDisambiguatorModel')
    PARTIAL_AverageEmbeddings = JslAnnoId('PARTIAL_AverageEmbeddings')
    PARTIAL_EntityChunkEmbeddings = JslAnnoId('PARTIAL_EntityChunkEmbeddings')
    PARTIAL_ChunkMergeApproach = JslAnnoId('PARTIAL_ChunkMergeApproach')
    PARTIAL_ChunkMergeApproach = JslAnnoId('PARTIAL_ChunkMergeApproach')
    PARTIAL_IOBTagger = JslAnnoId('PARTIAL_IOBTagger')
    PARTIAL_NerChunker = JslAnnoId('PARTIAL_NerChunker')
    PARTIAL_NerConverterInternalModel = JslAnnoId('PARTIAL_NerConverterInternalModel')
    PARTIAL_DateNormalizer = JslAnnoId('PARTIAL_DateNormalizer')
    PARTIAL_PosologyREModel = JslAnnoId('PARTIAL_PosologyREModel')
    PARTIAL_RENerChunksFilter = JslAnnoId('PARTIAL_RENerChunksFilter')
    PARTIAL_ResolverMerger = JslAnnoId('PARTIAL_ResolverMerger')
    PARTIAL_AnnotationMerger = JslAnnoId('PARTIAL_AnnotationMerger')
    PARTIAL_Router = JslAnnoId('PARTIAL_Router')
    PARTIAL_Word2VecApproach = JslAnnoId('PARTIAL_Word2VecApproach')
    PARTIAL_Word2VecApproach = JslAnnoId('PARTIAL_Word2VecApproach')
    PARTIAL_WordEmbeddings = JslAnnoId('PARTIAL_WordEmbeddings')
    PARTIAL_EntityRulerApproach = JslAnnoId('PARTIAL_EntityRulerApproach')
    PARTIAL_EntityRulerApproach = JslAnnoId('PARTIAL_EntityRulerApproach')
    PARTIAL_EntityRulerModel = JslAnnoId('PARTIAL_EntityRulerModel')
    PARTIAL_TextMatcherModel = JslAnnoId('PARTIAL_TextMatcherModel')
    PARTIAL_BigTextMatcher = JslAnnoId('PARTIAL_BigTextMatcher')
    PARTIAL_BigTextMatcherModel = JslAnnoId('PARTIAL_BigTextMatcherModel')
    PARTIAL_DateMatcher = JslAnnoId('PARTIAL_DateMatcher')
    PARTIAL_MultiDateMatcher = JslAnnoId('PARTIAL_MultiDateMatcher')
    PARTIAL_RegexMatcher = JslAnnoId('PARTIAL_RegexMatcher')
    PARTIAL_TextMatcher = JslAnnoId('PARTIAL_TextMatcher')
    PARTIAL_NerApproach = JslAnnoId('PARTIAL_NerApproach')
    PARTIAL_NerCrfApproach = JslAnnoId('PARTIAL_NerCrfApproach')
    PARTIAL_NerCrfApproach = JslAnnoId('PARTIAL_NerCrfApproach')
    PARTIAL_NerCrfApproach = JslAnnoId('PARTIAL_NerCrfApproach')
    PARTIAL_NerOverwriter = JslAnnoId('PARTIAL_NerOverwriter')
    PARTIAL_DependencyParserApproach = JslAnnoId('PARTIAL_DependencyParserApproach')
    PARTIAL_DependencyParserApproach = JslAnnoId('PARTIAL_DependencyParserApproach')
    PARTIAL_TypedDependencyParserApproach = JslAnnoId('PARTIAL_TypedDependencyParserApproach')
    PARTIAL_TypedDependencyParserApproach = JslAnnoId('PARTIAL_TypedDependencyParserApproach')
    PARTIAL_SentenceDetectorDLApproach = JslAnnoId('PARTIAL_SentenceDetectorDLApproach')
    PARTIAL_SentenceDetectorDLApproach = JslAnnoId('PARTIAL_SentenceDetectorDLApproach')
    PARTIAL_SentimentDetector = JslAnnoId('PARTIAL_SentimentDetector')
    PARTIAL_ViveknSentimentApproach = JslAnnoId('PARTIAL_ViveknSentimentApproach')
    PARTIAL_ViveknSentimentApproach = JslAnnoId('PARTIAL_ViveknSentimentApproach')
    PARTIAL_ContextSpellCheckerApproach = JslAnnoId('PARTIAL_ContextSpellCheckerApproach')
    PARTIAL_ContextSpellCheckerApproach = JslAnnoId('PARTIAL_ContextSpellCheckerApproach')
    PARTIAL_NorvigSweetingApproach = JslAnnoId('PARTIAL_NorvigSweetingApproach')
    PARTIAL_NorvigSweetingApproach = JslAnnoId('PARTIAL_NorvigSweetingApproach')
    PARTIAL_SymmetricDeleteApproach = JslAnnoId('PARTIAL_SymmetricDeleteApproach')
    PARTIAL_SymmetricDeleteApproach = JslAnnoId('PARTIAL_SymmetricDeleteApproach')
    PARTIAL_ChunkTokenizer = JslAnnoId('PARTIAL_ChunkTokenizer')
    PARTIAL_ChunkTokenizerModel = JslAnnoId('PARTIAL_ChunkTokenizerModel')
    PARTIAL_RecursiveTokenizer = JslAnnoId('PARTIAL_RecursiveTokenizer')
    PARTIAL_RecursiveTokenizerModel = JslAnnoId('PARTIAL_RecursiveTokenizerModel')
    PARTIAL_Token2Chunk = JslAnnoId('PARTIAL_Token2Chunk')
    PARTIAL_WordSegmenterApproach = JslAnnoId('PARTIAL_WordSegmenterApproach')
    PARTIAL_WordSegmenterApproach = JslAnnoId('PARTIAL_WordSegmenterApproach')
    PARTIAL_GraphExtraction = JslAnnoId('PARTIAL_GraphExtraction')
    PARTIAL_Lemmatizer = JslAnnoId('PARTIAL_Lemmatizer')
    PARTIAL_Normalizer = JslAnnoId('PARTIAL_Normalizer')
    PARTIAL_SpanBertCorefModel = JslAnnoId('PARTIAL_SpanBertCorefModel')







class NLP_HC_NODE_IDS:  # or Mode Node?
    """All avaiable Feature nodes in Healthcare Library.
    Defines High Level Identifiers

    Used to cast the pipeline dependency resolution algorithm into an abstract grpah
    """
    ASSERTION_DL = JslAnnoId('assertion_dl')
    TRAINABLE_ASSERTION_DL = JslAnnoId('trainable_assertion_dl')
    ASSERTION_FILTERER = JslAnnoId('assertion_filterer')  # TODO traianble?
    ASSERTION_LOG_REG = JslAnnoId('assertion_log_reg')
    TRAINABLE_ASSERTION_LOG_REG = JslAnnoId('trainable_assertion_log_reg')
    CHUNK2TOKEN = JslAnnoId('chunk2token')
    CHUNK_ENTITY_RESOLVER = JslAnnoId('chunk_entity_resolver')
    TRAINABLE_CHUNK_ENTITY_RESOLVER = JslAnnoId('traianble_chunk_entity_resolver')
    CHUNK_FILTERER = JslAnnoId('chunk_filterer')
    TRAINABLE_CHUNK_FILTERER = JslAnnoId('trainable_chunk_filterer')  # Todo not integrated
    CHUNK_KEY_PHRASE_EXTRACTION = JslAnnoId('chunk_key_phrase_extraction')
    CHUNK_MERGE = JslAnnoId('chunk_merge')
    TRAINABLE_CHUNK_MERGE = JslAnnoId('trainable_chunk_merge')  # Todo not integrated
    CONTEXTUAL_PARSER = JslAnnoId('contextual_parser')
    TRAIANBLE_CONTEXTUAL_PARSER = JslAnnoId('trainable_contextual_parser') # Todo not integrated
    DE_IDENTIFICATION = JslAnnoId('de_identification')
    TRAINABLE_DE_IDENTIFICATION = JslAnnoId('trainable_de_identification')
    DOCUMENT_LOG_REG_CLASSIFIER = JslAnnoId('document_log_reg_classifier')
    TRAINABLE_DOCUMENT_LOG_REG_CLASSIFIER = JslAnnoId('traianble_document_log_reg_classifier')
    DRUG_NORMALIZER = JslAnnoId('drug_normalizer')
    FEATURES_ASSEMBLER = JslAnnoId('features_assembler')
    GENERIC_CLASSIFIER = JslAnnoId('generic_classifier')
    TRAINABLE_GENERIC_CLASSIFIER = JslAnnoId('traianble_generic_classifier')
    IOB_TAGGER = JslAnnoId('iob_tagger')
    MEDICAL_NER = JslAnnoId('medical_ner')
    TRAINABLE_MEDICAL_NER = JslAnnoId('trainable_medical_ner')
    NER_CHUNKER = JslAnnoId('ner_chunker')
    NER_CONVERTER_INTERNAL = JslAnnoId('ner_converter_internal')
    NER_DISAMBIGUATOR = JslAnnoId('ner_disambiguator')
    TRAINABLE_NER_DISAMBIGUATOR = JslAnnoId('trainable_ner_disambiguator')
    RELATION_NER_CHUNKS_FILTERER = JslAnnoId('relation_ner_chunks_filterer')
    RE_IDENTIFICATION = JslAnnoId('re_identification')
    RELATION_EXTRACTION = JslAnnoId('relation_extraction')
    TRAINABLE_RELATION_EXTRACTION = JslAnnoId('trainable_relation_extraction')
    ZERO_SHOT_RELATION_EXTRACTION = JslAnnoId('zero_shot_relation_extraction')
    ZERO_SHOT_NER = JslAnnoId('zero_shot_ner')
    RELATION_EXTRACTION_DL = JslAnnoId('relation_extraction_dl')
    CHUNK_MAPPER_MODEL = JslAnnoId('chunk_mapper_model')

    # TRAINABLE_RELATION_EXTRACTION_DL = JslAnnoId('trainable_relation_extraction_dl')
    SENTENCE_ENTITY_RESOLVER = JslAnnoId('sentence_entity_resolver')
    TRAINABLE_SENTENCE_ENTITY_RESOLVER = JslAnnoId('trainable_sentence_entity_resolver')
    MEDICAL_BERT_FOR_TOKEN_CLASSIFICATION = JslAnnoId('medical_bert_for_token_classification')

    MEDICAL_BERT_FOR_SEQUENCE_CLASSIFICATION = JslAnnoId('medical_bert_for_sequence_classification')
    MEDICAL_DISTILBERT_FOR_SEQUENCE_CLASSIFICATION = JslAnnoId('medical_distilbert_for_sequence_classification')
    ENTITY_CHUNK_EMBEDDING = JslAnnoId('entity_chunk_embedding')
    MEDICAL_SUMMARIZER = JslAnnoId('med_summarizer')
    MEDICAL_TEXT_GENERATOR = JslAnnoId('med_text_generator')
    MEDICAL_QUESTION_ANSWERING = JslAnnoId('med_question_answering')

class OCR_NODE_IDS:
    """All available Feature nodes in OCR
    Used to cast the pipeline dependency resolution algorithm into an abstract graph
    """
    # Visual Document Understanding
    VISUAL_DOCUMENT_CLASSIFIER = JslAnnoId('visual_document_classifier')
    VISUAL_DOCUMENT_NER = JslAnnoId('visual_document_NER')

    # Object Detection
    IMAGE_HANDWRITTEN_DETECTOR = JslAnnoId('image_handwritten_detector')

    # TABLE Processors/Recognition
    IMAGE_TABLE_DETECTOR = JslAnnoId('image_table_detector')
    IMAGE_TABLE_CELL_DETECTOR = JslAnnoId('image_table_cell_detector')
    IMAGE_TABLE_CELL2TEXT_TABLE = JslAnnoId('image_table_cell2text_table')

    # PDF Processing
    PDF2TEXT = JslAnnoId('pdf2text')
    PDF2IMAGE = JslAnnoId('pdf2image')
    IMAGE2PDF = JslAnnoId('image2pdf')
    TEXT2PDF = JslAnnoId('text2pdf')
    PDF_ASSEMBLER = JslAnnoId('pdf_assembler')
    PDF_DRAW_REGIONS = JslAnnoId('pdf_draw_regions')
    PDF2TEXT_TABLE = JslAnnoId('pdf2table')

    # DOCX Processing
    DOC2TEXT = JslAnnoId('doc2text')
    DOC2TEXT_TABLE = JslAnnoId('doc2text_table')
    DOC2PDF = JslAnnoId('doc2pdf')
    PPT2TEXT_TABLE = JslAnnoId('ppt2text_table')
    PPT2PDF = JslAnnoId('ppt2pdf')

    # DICOM Processing
    DICOM2IMAGE = JslAnnoId('dicom2image')
    IMAGE2DICOM = JslAnnoId('IMAGE2DICOM')

    # Image Pre-Processing
    BINARY2IMAGE = JslAnnoId('binary2image')
    GPU_IMAGE_TRANSFORMER = JslAnnoId('GPU_IMAGE_TRANSFORMER')
    IMAGE_BINARIZER = JslAnnoId('image_binarizer')
    IMAGE_ADAPTIVE_BINARIZER = JslAnnoId('image_adaptive_binarizer')
    IMAGE_ADAPTIVE_THRESHOLDING = JslAnnoId('image_adaptive_thresholding')
    IMAGE_SCALER = JslAnnoId('image_scaler')
    IMAGE_ADAPTIVE_SCALER = JslAnnoId('image_adaptive_scaler')
    IMAGE_SKEW_CORRECTOR = JslAnnoId('image_skew_corrector')
    IMAGE_NOISE_SCORER = JslAnnoId('image_noise_scorer')
    IMAGE_REMOVE_OBJECTS = JslAnnoId('image_remove_objects')
    IMAGE_MORPHOLOGY_OPERATION = JslAnnoId('image_morphology_operation')
    IMAGE_CROPPER = JslAnnoId('image_cropper')
    IMAGE2REGION = JslAnnoId('image2region')
    IMAGE_LAYOUT_ANALZYER = JslAnnoId('image_layout_analyzer')
    IMAGE_SPLIT_REGIONS = JslAnnoId('image_split_regions')
    IMAGE_DRAW_REGIONS = JslAnnoId('image_draw_regions')

    # Character Recognition
    IMAGE2TEXT = JslAnnoId('image2text')
    IMAGE2TEXTPDF = JslAnnoId('image2textpdf')
    IMAGE2HOCR = JslAnnoId('image2hocr')
    IMAGE_BRANDS2TEXT = JslAnnoId('image_brands2text')

    # Other
    POSITION_FINDER = JslAnnoId('position_finder')
    UPDATE_TEXT_POSITION = JslAnnoId('update_text_position')
    FOUNDATION_ONE_REPORT_PARSER = JslAnnoId('foundation_one_report_parser')
    HOCR_DOCUMENT_ASSEMBLER = JslAnnoId('hocr_document_assembler')
    HOCR_TOKENIZER = JslAnnoId('hocr_tokenizer')




import streamlit as st
from sparknlp.annotator import *

from nlu.universe.component_universes import jsl_id_to_empty_component
from nlu.universe.feature_node_ids import NLP_NODE_IDS, NLP_HC_NODE_IDS
from nlu.universe.logic_universes import AnnoTypes
from nlu.universe.universes import Licenses


class EntityManifoldUtils:
    classifers_OS = [ClassifierDLModel, LanguageDetectorDL, MultiClassifierDLModel, NerDLModel, NerCrfModel,
                     YakeKeywordExtraction, PerceptronModel, SentimentDLModel,
                     SentimentDetectorModel, ViveknSentimentModel, DependencyParserModel, TypedDependencyParserModel,
                     T5Transformer, MarianTransformer, NerConverter]

    @staticmethod
    def insert_chunk_embedder_to_pipe_if_missing(pipe):

        """Scan component_list for chunk_embeddings. If missing, add new. Validate NER model_anno_obj is loaded"""
        # component_list.predict('Donald Trump and Angela Merkel love Berlin')

        classifier_cols = []
        has_ner = False
        has_chunk_embeds = True
        ner_component_names = ['named_entity_recognizer_dl', 'named_entity_recognizer_dl_healthcare']
        for c in pipe.components:
            if c.name == NLP_NODE_IDS.NER_DL or c.name == NLP_HC_NODE_IDS.MEDICAL_NER or c.type == AnnoTypes.TRANSFORMER_TOKEN_CLASSIFIER:
                has_ner = True
            if c.name == NLP_NODE_IDS.CHUNK_EMBEDDINGS_CONVERTER:
                return pipe
        if not has_ner:
            raise ValueError(
                "You Need to load a NER model_anno_obj or this visualization. Try nlu.load('ner').viz_streamlit_entity_embed_manifold(text)")

        ner_conveter_c, word_embed_c = None, None

        for c in pipe.components:
            if c.type == AnnoTypes.TOKEN_EMBEDDING:
                word_embed_c = c
            if c.name == NLP_NODE_IDS.NER_CONVERTER:
                ner_conveter_c = c
            if c.name == NLP_HC_NODE_IDS.NER_CONVERTER_INTERNAL:
                ner_conveter_c = c

        chunker = jsl_id_to_empty_component(NLP_NODE_IDS.CHUNK_EMBEDDINGS_CONVERTER)
        chunker.set_metadata(
            chunker.get_default_model(),
            'chunker', 'chunker', 'xx', False, Licenses.open_source)

        # chunker = embeddings_chunker.EmbeddingsChunker(nlu_ref='chunk_embeddings')

        chunker.model.setInputCols(ner_conveter_c.spark_output_column_names + word_embed_c.spark_output_column_names)
        chunker.model.setOutputCol('chunk_embedding')
        chunker.spark_input_column_names = ner_conveter_c.spark_output_column_names + word_embed_c.spark_output_column_names
        chunker.spark_output_column_names = ['chunk_embedding']
        # chunker.inputs = ner_conveter_c.spark_output_column_names + word_embed_c.spark_output_column_names
        # chunker.out_types = ['chunk_embedding']

        pipe.components.append(chunker)
        pipe.is_fitted = False
        pipe.fit()

        return pipe

    @staticmethod
    def find_chunk_embed_col(df):
        return 'chunk_embedding_chunker'
        for c in df.columns:
            ss = set(c.split('_'))
            if 'chunk' in ss: ss.remove('chunk')
            if 'embedding' in ss: ss.remove('embedding')
            if 'embeddings' in ss: ss.remove('embeddings')
            if len(ss) == 0: return c
        raise ValueError('Could not find chunk embed col')

    @staticmethod
    def get_ner_cols(df):
        """find NER pred, conf and class cols.
        Cols[0] =  entity_class_col
        Cols[1] =  entity_confidence
        Cols[2] =  entity_chunk

        """

        entity_class_col, entity_confidence_col, entity_chunk_col = None, None, None

        for c in df.columns:
            if 'entities' in c and 'class' in c:
                entity_class_col = c
            if 'entities' in c and 'confidence' in c:
                entity_confidence_col = c
            if 'entities' in c and 'confidence' not in c and 'confidence' not in c and 'origin' not in c:
                entity_chunk_col = c
        cols = [entity_class_col, entity_confidence_col, entity_chunk_col]
        if not any(cols):
            raise ValueError(f'Failure to resolve entities col for df cols = {df.columns}')
        return cols

    @staticmethod
    def find_entity_embed_col_pd(df, search_multi=False):
        """Find col that contains embed in pandas df """
        if not search_multi:
            for c in df.columns:
                if 'embed_entitiy' in c: return c
        else:
            e_cols = []
            for c in df.columns:
                if 'embed' in c: e_cols.append(c)
        return e_cols

    @staticmethod
    def find_embed_component(p):
        """Find first embed  component_to_resolve in component_list"""
        for c in p.components:
            if 'embed' in c.out_types[0]: return c
        st.warning("No Embed model_anno_obj in component_list")
        return None



from sparknlp.annotator import *

class NorvigSpellChecker:
    @staticmethod
    def get_default_model():
        return NorvigSweetingModel.pretrained() \
                   .setInputCols(["token"]) \
                   .setOutputCol("checked")

    @staticmethod
    def get_pretrained_model(name, language, bucket=None):
        return NorvigSweetingModel.pretrained(name,language,bucket) \
            .setInputCols(["token"]) \
            .setOutputCol("checked")

    @staticmethod
    def get_default_trainable_model():
        return NorvigSweetingApproach() \
            .setInputCols(["token"]) \
            .setOutputCol("checked") \
            .setDictionary("coca2017.txt", "[a-zA-Z]+")
